# coding: utf-8
""" As this class is a :class:`~base.DockerV` extension need to have the same values in the config
dictionary, but inside a config_container key::

        "container_config": {
            "apt_install": [],
            "pip_install": [],
            "domain": "localhost",
            "command": "sleep 20",
            "env_vars": {
                "odoo_config_file": "/home/odoo/.openerp_serverrc",
                "odoo_home": "/home/odoo",
                "odoo_user": "odoo"
            },
            "image_name": "busybox",
            "mem_limit": "768m",
            "ports": {
                "8069": 8069,
                "8072": None
            },
            "remove_previous": True,
            "working_folder": "/path/to/docker_volumes",
            "volumes":{
                "filestore": "/home/odoo/.local/share/Odoo",
                "logs": "/var/log/supervisor",
                "ssh": "/home/odoo/.ssh",
                "tmp": "/tmp"
            }
        },
        "instance": {
            "action_type": "replace",
            "task_id":"t1234",
            "customer_id": "customer80",
            "config": {
                "admin": "admin_pass",
                "admin_passwd": "123",
                "db_host": "172.17.42.1",
                "db_name": "{{ instance.task_id }}_{{ instance.customer_id }}",
                "db_password": "psql_user",
                "db_user": "psql_pass",
                "dbfilter": "{{ instance.task_id }}_{{ instance.customer_id }}.*",
                "list_db": True
            },
            "ssh_key":"base64_encoded_key",
            "repositories": [
                {
                    "branch": "8.0",
                    "commit": "8f159f276e7c8da4e5475e604e9057dcaadfedf9",
                    "depth": 1,
                    "is_dirty": False,
                    "name": "odoo",
                    "path": "odoo",
                    "repo_url": {
                        "origin": "git@github.com:Vauxoo/odoo.git"
                    },
                    "type": "git"
                },
            ],
            "update_module":"all",
            "install_module": None
        }

.. note:: For more information about *container_config* keys and values check
    :class:`~deployv.base.dockerv.DockerV` documentation

Where:

* *action_type*: Replace/create Odoo container
* *task_id*: associated task/issue/user story in the instance un the format [tiu]id
* *customer_id*: Unique customer id
* *config*: Dict with the values to be changed inside Odoo configuration file
* *ssh_key*: the id_rsa key ot be used to clone private repos, must be in base64 format
* *repositories*: A list of repos to be updated/added.
  Check `backupws <https://github.com/ruiztulio/backupws/blob/master/branches.py>`_
  if you want more detail about the format

"""
import tarfile
import logging
import github
from os import path, mkdir
from uuid import uuid4
import shutil
import time
import psycopg2
from psycopg2 import sql as psycopg2_sql
from docker.errors import APIError, ContainerError
import simplejson as json
import deployv_static
from passlib.context import CryptContext
from deployv.base.dockerv import DockerV
from deployv.base.postgresv import PostgresShell, PostgresConnector
from deployv.base import errors
from deployv.helpers import utils, container, database_helper, json_helper, pg_anon
from deployv.instance import ODOO_BINARY, PSQL_VERSIONS, TRAVIS_VERSIONS
from datetime import datetime
from jinja2 import Template


logger = logging.getLogger(__name__)  # pylint: disable=C0103
INSTANCE_TYPES = ['test', 'develop', 'updates', 'production']
ODOO_STAGES = {"production": "production", "updates": "staging",
               "develop": "dev", "test": "staging"}
UPDATE = ['develop', 'updates']
DEACTIVATE = ['test', 'develop', 'updates']
INSTALL_RIBBON = ['test', 'develop', 'updates']


class InstanceV(DockerV):

    def __init__(self, config, timeout=300, docker_url="unix://var/run/docker.sock"):
        self.__full_config = config
        self._skip_install_test_ribbon = False
        container_config = config.container_config
        logger.debug('Container config : %s',
                     json.dumps(container_config, sort_keys=True, indent=2))
        pg_container = '{prefix}_postgres'.format(prefix=config.prefix)
        container_config.update({'postgres_container_name': pg_container})
        if not container_config.get('container_name', False):
            container_config.update({
                'container_name': '{prefix}_odoo'.format(prefix=config.prefix)})
            env_vars = config.instance_config.get('config').copy()
            if not env_vars.get('admin'):
                env_vars.update({'admin': utils.random_string(6)})
            if not env_vars.get('admin_passwd'):
                env_vars.update({'admin_passwd': utils.random_string(6)})
            instance_type = config.instance_config.get('instance_type', 'production')
            env_vars.update({'instance_type': instance_type,
                             'odoo_stage': ODOO_STAGES.get(instance_type),
                             'customer': config.instance_config.get('customer_id')})
            self.__full_config.container_config.get('env_vars').update(env_vars)
        super(InstanceV, self).__init__(container_config,
                                        timeout=timeout, docker_url=docker_url)

    def start_odoo_container(self):
        """ Specifically starts an Odoo container and configure all the parameters,
        see :func:`~dockerv.DockerV.basic_info` for more information about the format and content

        :return: A dict with the basic info of container
        """
        logger.debug('Psql in container: %s', self.config.get('postgres_container', False))
        # if it will be a full stack container all the parameters are changed (no matter what is
        # specified in the config) to target the local env and added the nginx, postgres and ssh
        # ports
        if self.config.get('full_stack', False):
            self.config.get('ports').update({'22': None})
        self.pull()
        self.start_postgres_container()
        logger.info('Starting Odoo container %s',
                    self.config.get('container_name'))
        self.config.update({'container_hostname': container.generate_hostname(self.__full_config)})
        info = self.deploy_container()
        time.sleep(10)
        if self.config.get('full_stack', False):
            self.add_authorized_keys()
            supervisor_status = self.check_supervisor()
            if not supervisor_status:
                return {'error': 'Supervisor process could not start'}
            postgres_connection = self.check_postgres()
            if not postgres_connection.get('success'):
                return postgres_connection
        self.set_global_env_vars()
        # Add the PGHOST and PGPORT env vars here because we only know which values to set after
        # the container is created, mainly because we don't know if the postgres is in
        # another container.
        self.exec_cmd(
            "/bin/bash -c 'echo export PGHOST={db_host} | tee -a /home/odoo/.profile'"
            .format(db_host=self.config.get('env_vars').get('db_host')))
        self.exec_cmd(
            "/bin/bash -c 'echo export PGPORT={db_port} | tee -a /home/odoo/.profile'"
            .format(db_port=self.config.get('env_vars').get('db_port')))
        logger.info('Odoo container "%s" created. Ports %s',
                    info.get('name'), info.get('ports'))
        return info

    def check_supervisor(self):
        retry = 0
        supervisor_process = self.exec_cmd('pgrep supervisord')
        if supervisor_process:
            return True
        while True:
            retry += 1
            chown_process = self.exec_cmd('pgrep chown')
            if chown_process:
                logger.debug('chown command is still running, retrying %s', retry)
                time.sleep(3)
                continue
            supervisor_process = self.exec_cmd('pgrep supervisord')
            if supervisor_process:
                return True
            return False

    def check_postgres(self):
        res = {'success': True}
        retry = 0
        config = self.db_config
        config.update({'db_name': 'template1'})
        while retry <= 3:
            logger.info('Testing postgres connection')
            try:
                conn = PostgresConnector(utils.odoo2postgres(config))
                conn.check_config()
                logger.info('Postgres connection test passed')
                return res
            except psycopg2.OperationalError as error:
                if 'the database system is starting up' in error.message:
                    retry += 1
                    logger.warning('Postgres is starting up, retry %s', retry)
                    time.sleep(5)
                    continue
                logger.error(
                    'Could not connect to the postgres server: %s',
                    error.message)
                res.update({'success': False, 'error': error.message})
                return res

    def add_authorized_keys(self):
        tmp = self.__full_config.temp_folder or '/tmp'
        git_cli = github.Github()
        users = self.__full_config.instance_config.get('authorized_users', [])
        authorized_keys_path = path.join(tmp, 'authorized_keys')
        authorized_keys_tar_path = path.join(tmp, 'keys.tar')
        with open(authorized_keys_path, 'w') as authorized_keys:
            for user in users:
                user_keys = []
                try:
                    user_keys = git_cli.get_user(user).get_keys()
                except github.GithubException as error:
                    error_msg = error.data.get('message')
                    if 'Not Found' in error_msg:
                        logger.warning('User not found %s, could not retrieve any key', user)
                    else:
                        logger.warning(error_msg)
                for user_key in user_keys:
                    authorized_keys.write(user_key.key + '\n')
        with tarfile.open(authorized_keys_tar_path, 'w') as tar_keys:
            tar_keys.add(authorized_keys_path, arcname='authorized_keys')
        with open(authorized_keys_tar_path, 'r') as tar_data:
            self.cli.put_archive(self.config.get('container_name'),
                                 '/home/odoo/.ssh/', tar_data.read())
            self.exec_cmd('chown odoo:odoo /home/odoo/.ssh/authorized_keys')
            self.exec_cmd('cp /home/odoo/.ssh/authorized_keys /root/.ssh/')
            self.exec_cmd('chown root:root /root/.ssh/authorized_keys')
        utils.clean_files([authorized_keys_tar_path, authorized_keys_path])

    @property
    def is_running(self):
        """ If the Odoo instance ins running inside the container (or at least supervisord
        detects it) return True, False otherwise. If an unexpected state is detected raises
        a SupervisorStatusError exception

        :return: True or False
        """
        retry_triggers = [
            'refused connection', 'no such file', 'STARTING', 'STOPPING', 'BACKOFF']
        retry = 0
        while retry <= 3:
            try:
                res = self.exec_cmd('supervisorctl status ' + self.supervisor_app)
            except errors.NotRunning:
                retry += 1
                logger.debug('Container not running, retry %s', retry)
                time.sleep(4)
                continue
            logger.debug('is_running: %s', res.strip())
            if 'RUNNING' in res:
                return True
            if 'STOPPED' in res:
                return False
            if retry < 3 and any([msg for msg in retry_triggers if msg in res or res == '']):
                retry += 1
                logger.debug('The Odoo process is in an intermediate state or'
                             ' is not running yet, retrying %s', retry)
                time.sleep(4)
            else:
                raise errors.SupervisorStatusError('Unknown state: %s' % res)

    @property
    def instance_type(self):
        """ The instance type

        :return: Instance type: production, test, develop or false if it not set
        """
        return self.__full_config.instance_config.get('instance_type') or\
            self.docker_env.get('instance_type', 'production')

    @property
    def temp_folder(self):
        """ This is a very important property because the tmp folder is used to share files between
            the container and the host, widely used along the code.
            The best way (so far) is iterating all the mounted volumes until we find the /tmp

        :return: Full host path where the tmp volume is mounted
        """
        res = None
        try:
            inspected = self.inspect()
        except errors.NoSuchContainer:
            return res
        volumes = inspected.get('Mounts')
        for volume in volumes:
            if volume.get("Destination") == '/tmp' and volume:
                res = volume.get("Source")
                break
        return res

    @property
    def odoo_binary(self):
        return ODOO_BINARY[self.config_variables.get('version')]

    @property
    def supervisor_group(self):
        ct = self.config_variables.get('container_type')
        if ct == 'MULTIODOO':
            return 'multiodoo:*'
        return 'odoo'

    @property
    def supervisor_app(self):
        ct = self.config_variables.get('container_type')
        if ct == 'MULTIODOO':
            return 'multiodoo:odoo'
        return 'odoo'

    def start_instance(self):
        """ Start an instance inside a docker container using supervisord

        :return: Supervisord response
        """
        res = 'RUNNING'
        if not self.is_running:
            res = self.exec_cmd('supervisorctl start ' + self.supervisor_group)
            logger.debug('start_instance: %s', res.strip())
        return res

    def stop_instance(self):
        """ Stop an instance inside a docker container using supervisord

        :return: Supervisord response
        """
        res = 'STOPPED'
        if self.is_running:
            res = self.exec_cmd('supervisorctl stop ' + self.supervisor_group)
            logger.info('stop_instance: %s', res.strip())
        return res

    def ensure_process_check(self, process, check, possible_exceptions=False, max_tries=False):
        """Runs a given process and ensures that the result is the one that you want. If it is
        not, the process will run as many times as specified in max_tries, or until the check
        is passed.

        :param process: The process you want to run until it checks out. Should be a lambda
            function that receives the instance as the only param and returns the result of the
            process to be validated.
        :type process: lambda
        :param check: The check method for the value returned from process. Should be a lambda
            function that receives the result of the process function as the only param, checks
            the value is the expected one, and returns a boolean with that result.
        :type check: lambda
        :param possible_exceptions: A set of possible exceptions (classes) that can be raised
            from the process.
        :type possible_exceptions: set
        :param max_tries: The amount of times the given process will run until the result of the
            process is the expected one.
        :type max_tries: int
        """
        ok = False
        tries = 0
        if not possible_exceptions:
            possible_exceptions = Exception
        while tries < (max_tries or 99) and not ok:
            tries += 1
            try:
                process_result = process(self)
            except possible_exceptions:  # pylint: disable=E0712,W0703
                ok = False
            else:
                ok = check(self, process_result)
            if not ok:
                time.sleep(3)
        return ok

    def ensure_instance_status(self, ensure_running, max_tries=False):
        """Starts/stops an instance and ensures is either running or not, checking as many
        times as specified in max_tries.

        :param ensure_running: If True, it will start and ensure that the instance is running.
            If False, it will stop and ensure that the instance is not running.
        :type ensure_running: bool
        :param max_tries: The amount of times it will check
        :type max_tries: int
        """
        supervisor = self.ensure_process_check(
            lambda ins: ins.exec_cmd('pgrep supervisord'),
            lambda ins, res: bool(res),
            False, max_tries)
        if not supervisor:
            return False
        command = self.ensure_process_check(
            lambda ins: (
                ins.start_instance() if ensure_running
                else ins.stop_instance()
            ),
            lambda ins, res: bool(res),
            errors.SupervisorStatusError, max_tries)
        if not command:
            return False
        status = self.ensure_process_check(
            lambda ins: ins.is_running,
            lambda ins, res: res is ensure_running,
            False, max_tries)
        return status

    def run_and_log(self, command, attachment_name=False, skip_summary=False):
        """ Executes an Odoo related command inside the container (i.e.: update, install) and
            returns the resumed log and full path to the generated log. See
            :func:`~dockerv.helpers.utils.resume_log` for detailed information about
            the return format

        :param command: Command line to be executed inside Odoo container
        :return: Resumed log and full path to the generated file
        """
        self.stop_instance()
        logger.debug('run_and_log: %s', command)
        res = self.exec_cmd(command)
        resume = utils.resume_log(res.split('\n'))
        self.start_instance()
        attachment_name = (
            attachment_name or 'cmd_{timestamp}.log'.format(timestamp=utils.get_strtime())
        )
        working_path = path.join(self.temp_folder, attachment_name)
        with open(working_path, 'w') as log_file:
            log_file.write(res)
        resume.update({'log_file': working_path})
        if not skip_summary:
            logger.info('+-- Critical errors %s', len(resume.get('critical')))
            logger.info('+-- Errors %s', len(resume.get('errors')))
            logger.info('+-- Import errors %s', len(resume.get('import_errors')))
            logger.info('+-- Warnings %s', len(resume.get('warnings')))
            logger.info('+-- Translation Warnings %s', len(resume.get('warnings_trans')))
        logger.info('Logger was saved to: "%s"', working_path)
        return resume

    def update_db(self, db_name, modules=False):
        """ Updates the database using one of the possible methods of updating.

        Possible methods:
            - cou: uses click-odoo-update to update the modules that changed in the db.
            - normal: updates the provided modules or the ones specified in the json.

        :param db_name: Database name to be updated
        :param modules: List of modules to update, only used by the normal method.
        :return: Resumed log and full path to the generated file
        """
        if not self.__full_config.update_method:
            self.__full_config.update_method = 'normal'
        update_methods = {
            'cou': self._cou_update,
            'normal': self._module_update
        }
        if self.__full_config.update_method not in update_methods:
            return {'error': 'Unknown update method: %s' % self.__full_config.update_method}
        update_method = update_methods[self.__full_config.update_method]
        res = update_method(db_name)
        return res

    def _module_update(self, db_name):
        """ Updates the specified database by executing -u module for each module in
        the json.

        :param db_name: Name of the database to update.
        :return: dict with the logs of the update and the summary of the results.
        """
        res = {'result': {'update_logs': []}, 'attachments': []}
        image_name = self.config.get('image_name') or self.basic_info.get('image_name')
        modules = self.__full_config.instance_config.get('update_module') or []
        logger.info('Updating database %s with the image %s, modules %s',
                    db_name, image_name, modules)
        working_folder = path.join(self.config.get("working_folder"),
                                   self.config.get('container_name'))
        db_config = self.db_config
        update_command = (
            "/entry_point.py run --user odoo '/home/odoo/instance/odoo/{odoo}"
            " -c /home/odoo/.openerp_serverrc"
            " -d {database} -u {{module}} --logfile {{log}} --stop-after-init'"
            .format(odoo=self.odoo_binary, database=db_name))
        pg_shell = PostgresShell(utils.odoo2postgres(db_config))
        for module in modules:
            for attempt in range(2):
                log_file = 'update_{mod}_{date}.log'.format(mod=module, date=utils.get_strtime())
                log_path = path.join('/tmp/deployvlogs/', log_file)
                local_path = path.expanduser(path.join(working_folder, log_path.strip('/')))
                command = update_command.format(module=module, log=log_path)
                logger.info('Updating "%s"', module)
                # Terminate connections before updating
                pg_shell.terminate_sessions(db_name, db_config.get('db_owner'),
                                            db_config.get('db_owner_passwd'))
                try:
                    self.exec_cmd_image(command, self.config.get('volumes'),
                                        working_folder, image_name)
                except (ContainerError, APIError) as error:
                    res.update({'error': 'Error updating "{mod}": {err}'.format(
                        mod=module, err=utils.get_error_message(error))})
                    parsed_log = self._parse_update_log(local_path, module)
                    summary = parsed_log.get('summary')
                    if summary:
                        res['result']['update_logs'].append(summary)
                        res['attachments'].append(parsed_log.get('attachment'))
                    return res
                logger.info('Log was saved to "%s"', local_path)
                parsed_log = self._parse_update_log(local_path, module)
                if parsed_log.get('error') and not parsed_log.get('summary'):
                    # Return the error message if the log file couldn't be generated
                    res.update(parsed_log)
                    return res
                summary = parsed_log.get('summary')
                if parsed_log.get('error'):
                    if attempt:
                        res.update({'error': parsed_log.get('error')})
                        res['result']['update_logs'].append(summary)
                        res['attachments'].append(parsed_log.get('attachment'))
                        return res
                    logger.info('An error occured, trying again')
                    continue
                res['result']['update_logs'].append(summary)
                res['attachments'].append(parsed_log.get('attachment'))
                break
        return res

    def _cou_update(self, db_name):
        """ Updates the specified database by executing the click-odoo-update script
        from the OCA, which automatically detects the modules that changed and updates them.

        :param db_name: Name of the database to update.
        :return: dict with the logs of the update and the summary of the results.
        """
        res = {'result': {'update_logs': []}, 'attachments': []}
        image_name = self.config.get('image_name') or self.basic_info.get('image_name')
        working_folder = path.join(self.config.get("working_folder"),
                                   self.config.get('container_name'))
        logger.info('Updating database %s using COU', db_name)
        update_command = "/entry_point.py cou -d {database}".format(database=db_name)
        local_log_path = path.expanduser(
            path.join(working_folder, 'tmp/deployvlogs/cou_update.log'))
        pg_shell = PostgresShell(utils.odoo2postgres(self.db_config))
        for attempt in range(2):
            # Terminate connections before updating
            pg_shell.terminate_sessions(db_name, self.db_config.get('db_owner'),
                                        self.db_config.get('db_owner_passwd'))
            try:
                self.exec_cmd_image(update_command, self.config.get('volumes'),
                                    working_folder, image_name)
            except (ContainerError, APIError) as error:
                res.update({'error': 'Error updating using COU: {err}'.format(
                    err=utils.get_error_message(error))})
                parsed_log = self._parse_update_log(local_log_path, 'COU')
                summary = parsed_log.get('summary')
                if summary:
                    res['result']['update_logs'].append(summary)
                    res['attachments'].append(parsed_log.get('attachment'))
                return res
            logger.info('Log was saved to "%s"', local_log_path)
            parsed_log = self._parse_update_log(local_log_path, 'COU')
            if parsed_log.get('error') and not parsed_log.get('summary'):
                # Return the error message if the log file couldn't be generated
                res.update(parsed_log)
                return res
            summary = parsed_log.get('summary')
            if parsed_log.get('error'):
                if attempt:
                    res.update({'error': parsed_log.get('error')})
                    res['result']['update_logs'].append(summary)
                    res['attachments'].append(parsed_log.get('attachment'))
                    return res
                logger.info('An error occured, trying again')
                continue
            res['result']['update_logs'].append(summary)
            res['attachments'].append(parsed_log.get('attachment'))
            break
        return res

    def _parse_update_log(self, log_path, module):
        res = {}
        try:
            with open(log_path) as log_file:
                summary = utils.resume_log(log_file)
        except IOError as error:
            res.update({'error': 'Could not open log "{log}": {err}'.format(
                log=log_path, err=utils.get_error_message(error))
            })
            return res
        for log_type, log_num in sorted(summary.items(), key=lambda log: log[0]):
            logger.info('+-- %s %s', log_type.replace('_', ' ').title(), len(log_num))
        summary.update({'log_file': log_path})
        if summary.get('errors') or summary.get('critical'):
            update_error = 'An error ocurred while updating "{mod}"'.format(mod=module)
            res.update({'error': update_error})
            logger.error(update_error)
        res.update({
            'summary': summary,
            'attachment': {
                'file_name': path.basename(log_path),
                'file': utils.generate_attachment(log_path),
                'type': 'text/plain'
            }
        })
        return res

    def install_module(self, modules, db_name, without_demo=True):
        """ Executes "-i module" or "-i module1,module2" on the selected database and returns
            resumed log and full path to the generated log. See
            :func:`~dockerv.helpers.utils.resume_log` for detailed
            information about the return format

        :param modules: Module (str) o modules (list) to be installed
        :param db_name: Database name to be updated
        :param without_demo: Install the mdule/s with or without demo,
            if true will use --without-demo=all, if false omit the option,
            otherwise the module list
        :return: Resumed log and full path to the generated file
        """
        if utils.is_iterable(modules):
            modules_str = ','.join(modules)
        else:
            modules_str = modules

        if isinstance(without_demo, bool):
            wodemo = '--without-demo=all' if without_demo else ''
        elif utils.is_iterable(modules):
            wodemo = '--without-demo={}'.format(','.join(modules))
        elif isinstance(without_demo, str):
            wodemo = '--without-demo={}'.format(modules)
        logger.info('Installing modules %s in database %s',
                    modules_str, db_name)
        command_odoo = self.odoo_binary
        if self.config.get('env_vars'):
            env_vars = self.config.get('env_vars')
            if env_vars.get('odoo_home') and env_vars.get('odoo_user'):
                odoo_home = env_vars.get('odoo_home')
                odoo_user = env_vars.get('odoo_user')
        else:
            odoo_home = self.docker_env.get('odoo_home')
            odoo_user = self.docker_env.get('odoo_user')

        update_command = ('su {user} -c "{home}/instance/odoo/{odoo}'
                          ' -d {database} -i {modules} --stop-after-init {wodemo}"') \
            .format(home=odoo_home, user=odoo_user, database=db_name, modules=modules_str,
                    wodemo=wodemo, odoo=command_odoo)
        skip_modules = ['web_environment_ribbon_isolated']
        skip_summary = all(module in skip_modules for module in modules_str.split(','))
        attachment_name = 'install_{mod}_{timestamp}.log'.format(
            timestamp=utils.get_strtime(), mod=modules_str)
        return self.run_and_log(update_command, attachment_name, skip_summary)

    def move_file(self, filename):
        """Moves a file into the container temp shared volume

        :filename: File to be moved
        """
        actual_path = path.dirname(path.realpath(__file__))
        file_lib_src = path.join(
            actual_path, '..', 'helpers', filename)
        file_lib_dst = path.join(self.temp_folder, filename)
        shutil.copyfile(file_lib_src, file_lib_dst)

    def start_postgres_container(self):
        """ Specifically starts an Postgres container and configure users and connections,
        see :func:`~dockerv.DockerV.basic_info` for more information about the format and content

        :return: A dict with the basic info of container
        """
        if (not self.__full_config.container_config.get('postgres_container') or
                self.__full_config.backup_type in ['btrfs', 'zfs']):
            return {}
        destination = '/var/log/pg_log'
        if self.config.get('keep_db') or self.__full_config.reloading_config:
            info = self.inspect(self.config.get('postgres_container_name'))
            db_host = info.get('NetworkSettings').get('Gateway')
            self.config.update({'pg_log_postgres': self.get_specify_volumen(info, destination)})
            db_port = self.get_db_port(info)
            self.__full_config.instance_config.get("config").update({
                'db_host': db_host, 'db_port': db_port})
            self.config.get('env_vars').update({
                'db_port': db_port,
                'db_host': db_host, 'external_db_port': db_port})
            return {}
        ins_config = self.config.copy()
        psql_version = self.get_postgres_version()
        psql_image = "postgres:{version}".format(version=psql_version)
        role = 'createdb'
        if self.config.get('full_stack'):
            role = 'superuser'
            self.config.get('volumes').update({'pg_log': destination})
        self.config.update({"image_name": psql_image})
        self.pull()
        self.config.update({'ports': {'5432': None}})
        self.config.get('env_vars').update({
            'db_host': '127.0.0.1',
            'db_port': 5432,
            'postgres_password': 'postgres',
            'pg_log_path': destination
        })
        self.config.update({"container_name": self.config.get('postgres_container_name')})
        self.config.update({'container_hostname': container.generate_hostname(self.__full_config)})
        self.config.update({'command': ''})
        res = self.deploy_container()
        while 'accepting connections' not in self.exec_cmd('pg_isready', user='postgres'):
            time.sleep(3)
        cmd = "psql -c \"create user {user} with password '{password}' {role}\"".\
            format(user=self.db_config.get('db_user'), role=role,
                   password=self.db_config.get('db_password'))
        self.exec_cmd(cmd, user="postgres")
        # Create onwer user in postgreSQL
        cmd = "psql -c \"create user {user} with password '{password}' superuser\"".\
            format(user=self.db_config.get('db_owner'),
                   password=self.db_config.get('db_owner_passwd'))
        self.exec_cmd(cmd, user="postgres")
        info = self.inspect()
        self.config.update(ins_config)
        self.config.get('volumes').pop('pg_log', False)
        self.config.update({
            'pg_log_postgres': self.get_specify_volumen(info, destination)})
        db_host = info.get('NetworkSettings').get('Gateway')
        db_port = self.get_db_port(info)
        self.config.get('env_vars').update({
            'db_host': db_host, 'external_db_port': db_port, 'db_port': db_port})
        self.__full_config.instance_config.get('config').update({
            'db_host': db_host, 'db_port': db_port})
        # Don't expose the 5432 port in the odoo container
        # if a postgres container was created
        self.config.get('ports').pop('5432', False)
        return res

    def get_postgres_version(self):
        """Helper method used to obtain the postgres version required for this version of odoo.
        This makes it easier to get the information used for the btrfs/zfs requests and the
        information required to create the postgres containers.
        """
        try:
            image_data = self.inspect_image(self.config.get('image_name'))
            vars_list = image_data.get('Config').get('Env')
        except errors.NoSuchImage:
            vars_list = []
        image_vars = container.parse_env_vars(vars_list)
        psql_version = image_vars.get('psql_version', '9.6')
        return psql_version

    def get_specify_volumen(self, info, volume_path):
        """Get a specify volumen from the information of container.
        """
        source = [volume.get('Source') for volume in info.get('Mounts')
                  if volume.get('Destination') == volume_path]
        return source and source[0] or False

    def get_db_port(self, info_inspected):
        """Get the database port that use in the postgres container
        the instance to connect with postgres.

        :param info_inspected: The information of container.
        :type info_inspected: dict

        :return: the port that use the instance to connect with postgres.
        """
        ports = container.get_ports_dict(info_inspected)
        db_port = ports.get('5432')
        if isinstance(db_port, list):
            db_port = db_port[0].split(':')[1]
        return db_port

    def restore_database(self, db_name=False, backup_dir=False):
        res = {'result': {}}
        error = False
        if not self.docker_id:
            error = 'The container %s does not exist' % self.config.get('container_name')
        db_name_config = self.__full_config.instance_config.get('config').get('db_name')
        customer_id = (self.__full_config.instance_config.get('customer_id') or
                       self.docker_env and self.docker_env.get("customer"))
        if not customer_id:
            error = 'No such customer to restore the database'
        if error:
            return {'error': error}
        use_template = self.__full_config.container_config.get('use_template')
        jobs = self.__full_config.jobs or 5
        helper_obj = database_helper.DatabaseHelper.get_helper(use_template)
        db_helper = helper_obj(
            utils.odoo2postgres(self.db_config), self._get_admin_config(),
            self.__full_config.postgres.get('extensions'))
        backup_dir = backup_dir or self.__full_config.backup_src
        res_search = db_helper.search_candidate(backup_dir, customer_id)
        if not res_search[0]:
            return {'error': res_search[1]}
        res.get('result').update({'backup': path.basename(res_search[1])})
        try:
            candidate = utils.decompress_files(
                res_search[1], self.__full_config.temp_working_folder)
        except (EOFError, IOError) as error:
            res.update({'error': utils.get_error_message(error)})
            return res
        backup_size = utils.get_size(candidate)
        res.get('result').update({'backup': path.basename(candidate),
                                  'source': candidate, 'backup_size': backup_size})
        filestore = path.join(candidate, 'filestore')
        db_name = db_name or db_name_config or utils.generate_dbname(
            self.__full_config, path.basename(res_search[1]), self.__full_config.prefix)
        self.stop_instance()
        create_res = db_helper.create_database(candidate, db_name, self.db_config.get('db_owner'),
                                               self.db_config.get('db_owner_passwd'), jobs=jobs)
        if not create_res[0]:
            res.update({'error': create_res[1]})
            return res
        create_date = datetime.now().strftime('%Y-%m-%d %H:%M:%S')
        filestore_res = self.restore_filestore(filestore, db_name)
        if filestore_res.get('error'):
            return filestore_res
        res.get('result').update({'database_name': db_name, 'create_date': create_date})
        self.create_extension_unaccent(db_name)
        return res

    def files2backup(self, database_name, tmp_dir, data_dir=False):
        """ Generates the files that will be added to the backup, such as
        json with the information of the repositories used in that instance,
        filestore and db dump.

        :param database_name: Name of the database that will be dumped
        :param tmp_dir: Temporal directory where the dump will be stored
        :returns: List of the files that will be compressed in the backup
        """
        files = []
        jobs = self.__full_config.jobs
        postgres_cfg = self.db_config
        data_dir = data_dir or postgres_cfg.get('data_dir')
        postgres_cfg.update({'database': database_name})
        postgres = PostgresShell(utils.odoo2postgres(postgres_cfg))
        dump_base_name = 'dump.sql' if not jobs else 'database_dump'
        dump_name = path.join(tmp_dir, dump_base_name)
        exclude = self.__full_config.postgres.get("exclude_tables") or ""
        dump_name = postgres.dump(database_name, dump_name, jobs=jobs, exclude=exclude)
        if not dump_name:
            logger.error('Database could not be dumped')
            return files
        files.append(
            dump_name if not path.isdir(dump_name) else (dump_name, path.basename(dump_name))
        )
        try:
            manifest_path = self._get_db_manifest(tmp_dir)
            files.append(manifest_path)
        except errors.CommandError as error:
            error_msg = utils.get_error_message(error)
            # TODO: Only log an error when the container does not exist to provide
            # the same treatment as the filestore, this allows us to generate a
            # backup with the required files during the production release, where
            # the container doesn't exist but the files are obtained from another
            # place. We need to improve this logic in the future though.
            if 'Container does not exist' not in error_msg:
                raise
            logger.error('The container %s does not exist,'
                         ' manifest file won\'t be added to the backup',
                         self.docker_id or self.config.get('container_name'))
        if data_dir:
            attachments_folder = path.join(
                data_dir, 'filestore', postgres_cfg.get('database')
            )
            if path.exists(attachments_folder):
                logger.debug('Attachments folder "%s"', attachments_folder)
                files.append((attachments_folder, 'filestore'))
            else:
                logger.warning(
                    'Folder "%s" does not exists, attachments are not being '
                    'added to the backup', attachments_folder)
        else:
            logger.info('There is not attachments folder to backup')
        try:
            json_repos = self.save_repositories("repositories.json")
            if path.exists(json_repos):
                files.append((json_repos, 'repositories.json'))
        except errors.CommandError as error:
            error_msg = utils.get_error_message(error)
            logger.warning('Failed to generate the branches json: \n%s', error_msg)
        return files

    def compress_files(self, files, backup_name, destination, cformat):
        """ Creates a compressed file in the specified destination with the name,
        format, and files provided.

        :param files: List of  files that will be added to the compressed file
        :param backup_name: Name of the compressed file that will be created
        :param destination: Directory where the compressed file will be created
        :param cformat: Format used to compress the files (bz2, gz)
        :return: Name of the new compressed file
        """
        logger.info('Compressing files')
        logger.debug('Files : %s', str(files))
        full_name = utils.compress_files(
            backup_name, files, dest_folder=destination, cformat=cformat)
        logger.info('Compressed backup, cleaning')
        for element in files:
            if not hasattr(element, '__iter__'):
                utils.clean_files(element)
        return full_name

    def generate_backup(self, params):
        """ Generate an instance backup that contains:
            - Database dump in sql format
            - A folder with instance attachments called *filestore*
            - Json file with the repos ans hashes from the actual instance

        .. note:: If the method cannot access the filestore folder it won't be included in the
            backup

        :param dest_folder: Folder where the backup will be stored
        :param  reason: Optional parameter that is used in case
                          there is a particular reason for the backup
        :return: Full path to the generated backup
        """
        additional_files = params.get('additional_files') or []
        reason = params.get('reason') or self.__full_config.reason
        prefix = params.get('prefix') or self.__full_config.prefix
        cformat = params.get('cformat') or self.__full_config.cformat
        db_name = params.get('db_name')
        try:
            files = self.files2backup(
                db_name, self.__full_config.temp_working_folder, params.get('data_dir'))
        except errors.DumpError as error:
            logger.error('An error raise during the database backup, cleaning up')
            raise error  # pylint: disable=E0702
        files.extend(additional_files)
        bkp_name = utils.generate_backup_name(db_name, reason, prefix)
        res = self.return_backup(files, bkp_name, params.get('dest_folder'), cformat)
        return res

    def return_backup(self, files, bkp_name, dest_folder, cformat):
        if not cformat:
            return {'files': files, 'name': bkp_name,
                    'tmp_dir': self.__full_config.temp_working_folder}
        if cformat == 'folder':
            res = self._generate_decompressed_backup(
                files, bkp_name, dest_folder
            )
        else:
            res = self.compress_files(files, bkp_name, dest_folder, cformat)
        return res

    def _generate_decompressed_backup(self, files, name, destination):
        """Moves all the files obtained during the backup to a folder
        with the specified name in the destionation.

        :param files: list of paths to the files that will be added to the backup
        :param name: name of the new backup
        :param destination: directory where the new backup will be store
        """
        logger.info('Generating decompressed backup')
        tmp_name = path.join(destination, '._{}'.format(name))
        logger.info('Creating temp folder %s', tmp_name)
        mkdir(tmp_name)
        logger.info('Temp folder created')
        for file_path in files:
            is_iterable = utils.is_iterable(file_path)
            src = file_path if not is_iterable else file_path[0]
            dest = (path.join(tmp_name, path.basename(file_path)) if not is_iterable else
                    path.join(tmp_name, file_path[1]))
            logger.info('Adding file %s to %s', src, dest)
            if path.isdir(src):
                shutil.copytree(src, dest)
            else:
                shutil.copy(src, dest)
        res = path.join(destination, name)
        shutil.move(tmp_name, res)
        logger.info('Backup %s generated', res)
        return res

    def restore_filestore(self, src_folder, database_name):
        """ Restore a filestore folder into a docker container that is already running
            and has the /tmp folder mounted as a volume un the host

        :param src_folder: Full path to the folder that contains the filestore you want to restore
        :param database_name: Database name being restored
        """
        working_folder = path.join(self.config.get("working_folder"),
                                   self.config.get('container_name'))
        dest = path.join(path.expanduser(working_folder), 'filestore')
        volumes = {
            path.abspath(src_folder): {
                'bind': '/tmp/filestore',
                'ro': False,
            },
            dest: {
                'bind': self.config.get('volumes').get('filestore'),
                'ro': False
            }
        }
        bash_lines = [
            "rm -rf {fsname}",
            "mkdir -p /home/odoo/.local/share/Odoo/filestore/",
            "cp -R /tmp/filestore {fsname}",
            "chown -R odoo:odoo /home/odoo/.local",
            "chmod -R o+r {fsname}"
        ]
        res = self.execute_in_auxiliary_container(database_name, bash_lines, volumes)
        return res

    def execute_in_auxiliary_container(self, db_name, commands, volumes=False):
        """Method that creates an auxiliary container copy of the
        instance and executes the provided commands in a single docker
        run (the commands are separated by `&&`). This is so we can execute
        commands that modify data contained in the container's volumes that
        may require too many resources without taking the risk of killing
        the main container.

        :param db_name: Name of the database, use to know where the volumes
            created by deployv are.
        :type db: str
        :param commands: List of commands to be executed
        :type commands: list
        :return: Dictionary with the `result` key (list) if all the commands are
        executed, and/or the `error` key (list) if any or all of them failed.
        """
        res = {}
        results = []
        errors = []
        logger.debug('Getting properties')
        fs_name = path.join('/home/odoo/.local/share/Odoo/filestore/', db_name)
        logger.debug('Creating auxiliary container')
        for cmd in commands:
            cmd = cmd.format(fsname=fs_name)
            try:
                container_cfg = self.inspect()
            except errors.NoSuchContainer:
                container_cfg = {}
            image = container_cfg.get('Config', {}).get('Image') or self.config.get('image_name')
            cmd_res = self.exec_cmd_image("bash -c '%s'" % (cmd), volumes, image=image)
            if cmd_res.get('error'):
                errors.append(cmd_res['error'])
            elif cmd_res.get('result'):
                results.append(cmd_res['result'])
        if results:
            res.update({'result': '\n'.join(results)})
        if errors:
            res.update({'error': '\n'.join(errors)})
        return res

    def execute_query(self, db_config, query, allow_errors=False):
        """Executes the provided query upon the specified database.

        :param db_config: Name of the database where the query will be executed,
        :param query: SQL query to execute.
        :param allow_errors: If True, the errors raised regarding missing or already
            existing tables while executing the query will be ignored, a
            psycopg2.ProgrammingError will be raised otherwise.
        """
        allowed_errors = ([] if not allow_errors else
                          ['relation', 'does not exist', 'already exists',
                           'no existe', 'ya existe'])
        with PostgresConnector(db_config) as db:
            try:
                res = db.execute(query)
            except psycopg2.ProgrammingError as error:
                error_msg = utils.get_error_message(error)
                if any([string in error_msg for string in allowed_errors]):
                    logger.warning(
                        'An error ocurred, skipped. %s', error_msg.split('\n')[0])
                    return []
                raise
        return res

    def get_views(self, database):
        db_config = self.db_config
        db_config.update({'db_name': database})
        db_config = utils.odoo2postgres(db_config)
        arch_exists = self.execute_query(
            db_config,
            "select 1 from information_schema.columns where"
            " table_name='ir_ui_view' and column_name='arch'"
        )
        arch_column = 'arch' if arch_exists else 'arch_db'
        sql = """SELECT ir_model_data.module || '.' || ir_model_data.name xml_id, {arch} as arch
            FROM ir_model_data
            JOIN ir_ui_view ON res_id = ir_ui_view.id
            WHERE ir_model_data.model = 'ir.ui.view'
            ORDER BY xml_id;""".format(arch=arch_column)
        views = self.execute_query(db_config, sql)
        return views

    def compare_views(self, original_views, modified_views):
        """
        Compare all the views from views from production with the views from updates and returns
        a proper report

        :param original_views: The views from production database (a copy of course)
        :param modified_views: The views from the db with all changes applied
            (-u all, -u app_module).
        :return: a dict with the added, deleted and updated views. In the case of updated will
            return the diff between the production database and the updates database
        """
        res = {'updated': [], 'added': [], 'deleted': []}
        copy_orig_views = original_views[:]  # to maintain original_views unchanged
        total = len(copy_orig_views)
        for view_modified in modified_views:
            found = False
            ind = 0
            while not found and ind < total:
                if copy_orig_views[ind]['xml_id'] == view_modified['xml_id']:
                    if copy_orig_views[ind]['arch'] != view_modified['arch']:
                        res['updated'].append({
                            'xml_id': copy_orig_views[ind]['xml_id'],
                            'original': copy_orig_views[ind]['arch'],
                            'modified': view_modified['arch']
                        })
                    found = True
                    copy_orig_views.pop(ind)
                    total -= 1
                else:
                    ind += 1   # only in else, when no pop is done in copy_orig_views
            if not found:
                res['added'].append(view_modified)
        for original_view in copy_orig_views:
            res['deleted'].append(original_view)
        return res

    def get_translations(self, database):
        db_config = self.db_config
        db_config.update({'db_name': database})
        db_config = utils.odoo2postgres(db_config)
        translations = self.execute_query(
            db_config,
            """SELECT value,id,name,module FROM ir_translation"""
        )
        return translations

    def compare_translations(self, original_translations, modified_translations):
        """ Compare all the translated fields from two databases and returns a proper report

        :param original_translations: The translations contained in the production
                                      database (copy of course).
        :param modified_translations: The translations contained in the updates database with
                                      all the changes that will be applied in the
                                      production database.
        :return: A dict with the added, updated and removed translations between the production
                 database and the updates database.
        """
        res = {'updated': [], 'added': [], 'deleted': []}
        copy_orig_translations = original_translations[:]  # to maintain original_translations
        total = len(copy_orig_translations)
        for modified_translation in modified_translations:
            found = False
            ind = 0
            while not found and ind < total:
                if copy_orig_translations[ind]['id'] == modified_translation['id']:
                    if copy_orig_translations[ind]['value'] != modified_translation['value']:
                        res['updated'].append({
                            'name': copy_orig_translations[ind]['name'],
                            'module': copy_orig_translations[ind]['module'],
                            'original': copy_orig_translations[ind]['value'],
                            'modified': modified_translation['value']
                        })
                    found = True
                    copy_orig_translations.pop(ind)
                    total -= 1
                ind += 1
            if not found:
                res['added'].append({
                    'name': modified_translation['name'],
                    'module': modified_translation['module'],
                    'value': modified_translation['value']
                })
        for original_translation in copy_orig_translations:
            res['deleted'].append({
                'name': original_translation['name'],
                'module': original_translation['module'],
                'value': original_translation['value'],
            })
        return res

    def get_menus(self, database):
        """ Retrieves all the menus from the provided database

        :param database: Name of the database from where deployv will
                         retrieve all the menus.
        :return: Dictionary containing all the menus from the
                 provided database.
        """
        res = dict()
        db_config = self.db_config
        db_config.update({'db_name': database})
        db_config = utils.odoo2postgres(db_config)
        menu_sql = """SELECT ir_model_data.module || '.' || ir_model_data.name AS xml_id,
                        res_id, ir_ui_menu.name
                FROM ir_model_data
                JOIN ir_ui_menu ON res_id = ir_ui_menu.id
                WHERE ir_model_data.model = 'ir.ui.menu';"""

        tree_sql = """WITH RECURSIVE search_menu(id, parent_id, name, depth, hierarchypath) AS (
            SELECT menu.id, menu.parent_id, menu.name, 1, ppmenu.name || '->' || menu.name
            as hierarchypath
            FROM ir_ui_menu AS menu
            JOIN ir_ui_menu AS ppmenu
            ON menu.parent_id = ppmenu.id
            UNION ALL
                SELECT menu.id, menu.parent_id, menu.name, pmenu.depth + 1,
                    hierarchypath || '->' || menu.name
                FROM ir_ui_menu as menu
                JOIN search_menu as pmenu
                ON menu.parent_id = pmenu.id
            )
            SELECT * FROM search_menu WHERE id = {menu} ORDER BY depth DESC LIMIT 1;"""
        menus = self.execute_query(db_config, menu_sql)
        for menu in menus:
            tree = self.execute_query(db_config, tree_sql.format(menu=menu['res_id']))
            hierarchy = tree[0] if tree else {}
            menu.update({'hierarchypath': hierarchy.get('hierarchypath')})
            res.update({
                menu['xml_id']: menu,
            })
        return res

    def compare_menus(self, original_menus, modified_menus):
        """Compare all the menu items from two json values, obtained from backup database,
        and return a proper report

        :param original_menus: A list of dicts, containing info of original menu items.
        :param modified_menus: A list of dicts, containing info of updated menu items.
        :return: A dict with the added, updated and removed menus between both json values."""
        res = {'updated': [], 'added': [], 'deleted': []}
        for uxml_id in modified_menus:
            if uxml_id in original_menus:
                if original_menus[uxml_id]['name'] != modified_menus[uxml_id]['name']:
                    res['updated'].append({
                        'xml_id': uxml_id,
                        'original': original_menus[uxml_id]['name'],
                        'modified': modified_menus[uxml_id]['name'],
                        'hierarchypath': modified_menus[uxml_id].get('hierarchypath')
                    })
            else:
                res['added'].append({
                    'xml_id': uxml_id,
                    'name': modified_menus[uxml_id]['name'],
                    'hierarchypath': modified_menus[uxml_id].get('hierarchypath')
                })
        for pxml_id in original_menus.keys():
            if pxml_id not in modified_menus:
                res['deleted'].append({
                    'xml_id': original_menus[pxml_id]['xml_id'],
                    'name': original_menus[pxml_id]['name'],
                    'hierarchypath': original_menus[pxml_id].get('hierarchypath')
                })
        return res

    def get_fields(self, database):
        """Gets all the fields from the specified database and returns
            their basic information.

        :param database: name of the database from where the fields will be
            retrieved.
        :return: List of dicts with the information of the fields::
            .. code-block:: json

                    [
                        {
                            'model': model,
                            'name': field,
                            'description': description,
                            'type': field_type \
                        } \
                    ]
        """
        sql = """
        select model, name, field_description as description,
        ttype as type from ir_model_fields;
        """
        db_config = self.db_config
        db_config.update({'db_name': database})
        db_config = utils.odoo2postgres(db_config)
        res = self.execute_query(db_config, sql)
        return res

    def compare_fields(self, original_fields, modified_fields):
        """Compare all the field items from two json values, obtained from backup database,
        and return a proper report

        :param original_fields: A list of dicts, containing info of original fields items.
        :param modified_fields: A list of dicts, containing info of updated fields items.
        :return: A dict with the added, updated and removed fields between both json values."""
        res = {'updated': [], 'added': [], 'deleted': []}
        copy_orig_fields = original_fields[:]  # to maintain original_fields unchanged
        total = len(copy_orig_fields)
        for modified in modified_fields:
            found = False
            ind = 0
            while not found and ind < total:
                if modified['model'] == copy_orig_fields[ind]['model'] \
                        and modified['name'] == copy_orig_fields[ind]['name']:
                    if modified['type'] != copy_orig_fields[ind]['type'] \
                            or modified['description'] != copy_orig_fields[ind]['description']:
                        res['updated'].append({
                            'model': copy_orig_fields[ind]['model'],
                            'name': copy_orig_fields[ind]['name'],
                            'original': {'type': copy_orig_fields[ind]['type'],
                                         'description': copy_orig_fields[ind]['description']},
                            'modified': {'type': modified['type'],
                                         'description': modified['description']},
                        })
                    found = True
                    copy_orig_fields.pop(ind)
                    total -= 1
                ind += 1
            if not found:
                res['added'].append(modified)
        for original_field in copy_orig_fields:
            res['deleted'].append(original_field)
        return res

    def get_odoo_users(self, database, user_id=False):
        """ Retrieves the user ids of all the current users registered in
        the odoo instance.

        :param database: Name of the database where deployv will retrieve
                         the users ids from.
        :param user_id (optional): Id of the user to filter the query.
        :return: List of users ids
        """
        psql_dict = self.db_config
        psql_dict.update({'db_name': database})
        query = psycopg2_sql.SQL("SELECT id,login FROM res_users ")
        if user_id:
            query += psycopg2_sql.SQL("where id={0}").format(psycopg2_sql.Literal(user_id))
        with PostgresConnector(utils.odoo2postgres(psql_dict)) as db:
            users = db.execute(query.as_string(db._PostgresConnector__conn))
        return users

    @staticmethod
    def change_password(user_id, new_pass, db_name, db_config):
        """ Changes the specified user_id password

        :param user_id: Users id
        :param new_pass: New password that will be set to the user
        :param db_name: Database name to be updated
        :return: True if the password was changed, None otherwise
        """
        logger.info('Changing password for user_id %s', user_id)
        default_crypt_context = CryptContext(
            ['pbkdf2_sha512', 'md5_crypt'],
            deprecated=['md5_crypt'],
        )
        db_config.update({'db_name': db_name})
        with PostgresConnector(utils.odoo2postgres(db_config)) as db:
            try:
                res_user = db.execute(("select * from res_users where id = %(id)s"),
                                      {'id': user_id})
            except psycopg2.ProgrammingError as error:
                error_msg = ("Failed to change the password of the user {user}: {error}"
                             .format(user=user_id, error=str(error)))
                logger.error(error_msg)
                return {'error': error_msg}
            res = {"result": True}
            if res_user and res_user[0].get('password_crypt'):
                logger.info('Encrypted password')
                crypted_passwd = default_crypt_context.encrypt(new_pass)
                db.execute(("update res_users set password='',"
                            " password_crypt=%(passwd)s where id = %(id)s"),
                           {'passwd': crypted_passwd, 'id': user_id})
            elif res_user and res_user[0].get('password'):
                logger.info('Non encrypted password')
                db.execute("update res_users set password=%(passwd)s where id = %(id)s",
                           {'passwd': new_pass, 'id': user_id})
        res.update({"result": new_pass})
        return res

    def check_keys(self):
        """ Check if the provided keys into the docker container allows the user to connect
            github.com but does not check permissions

        :return: True if it is a valid key, False otherwise
        """
        logger.info('Checking keys in container %s', self.docker_id)
        res = self.exec_cmd('su {user} -c "ssh -T git@github.com"'
                            .format(user=self.config.get('env_vars').get('odoo_user')))
        if u'successfully authenticated, but GitHub does not provide shell access.' in res:
            logger.info('Keys checked properly')
            return True
        logger.error(
            'An error occurred while trying to connect to github.com: %s', res.strip())
        return False

    def install_deps(self):
        """ Install dependencies specified in the config json (if any). To see the return
        format check :func:`~deployv.base.dockerv.DockerV.install_packages`

        :return: A dic with the installation process
        """
        apt_list = self.config.get('apt_install', None)
        pip_list = self.config.get('pip_install', None)
        res = self.install_packages(apt_list, pip_list)
        return res

    def clean_volumes(self):
        """Removes all the files in the /tmp folder, .ssh directory, in the filestore,
        and the logs inside the container. This method is used to clean the files before
        destroying a container so they are not present the next time we deploy the same container
        """
        logger.info('Cleaning volumes')
        use_snapshot = self.__full_config.backup_type in ['btrfs', 'zfs']
        for key, volume in self.config.get('volumes').items():
            if (self.config.get('keep_db') or use_snapshot) and key == 'filestore':
                continue
            logger.info('Removing volume: %s', volume)
            self.exec_cmd('rm -rf {volume}'.format(volume=volume))

    def set_parameters(self, db_name, parameters):
        """Inserts the provided parameters in the ir.config_parameter
        table of the provided database.

        :param db_name: Name of the database where the parameters will be inserted
        :param parameters: Dictionary with the parameters that will be created in the db where
            each key in the dictionary is the name (key) of the parameter and the value of that key
            is the value of that parameter.
        :type parameters: dict
        """
        config = self.db_config
        config.update({'db_name': db_name})
        with PostgresConnector(utils.odoo2postgres(config)) as db:
            for key, value in parameters.items():
                date = datetime.now().strftime('%Y-%m-%d %H:%M:%S.%f')
                sql = (
                    "INSERT INTO ir_config_parameter ("
                    "    create_uid, write_date, value, write_uid, key, create_date)"
                    " VALUES (1, '{date}', '{value}', 1, '{key}', '{date}');"
                    .format(value=value, key=key, date=date))
                try:
                    db.execute(sql)
                except (psycopg2.IntegrityError, psycopg2.ProgrammingError) as error:
                    error_msg = utils.get_error_message(error)
                    if 'duplicate key value violates unique constraint' in error_msg:
                        # Remove parameter from the new db and insert it again with the new values
                        db.execute("delete from ir_config_parameter where key = '{key}'"
                                   .format(key=key))
                        db.execute(sql)
                    elif 'does not exist' in error_msg:
                        logger.warning('Failed to set the %s parameter, skipping', key)
                    else:
                        raise
        return True

    def set_global_env_vars(self, rcfile=False):
        """Adds the environment variables specified in the config json to the
            sh rc file so all users that connect to the instance can use them.

        :param rcfile: full path to the rc file where the env vars will be set
            (bash.bashrc, zshrc, etc), if no path is specified /etc/bash.bashrc will be used
        :type rcfile: str

        """
        env_vars = []
        if not rcfile:
            rcfile = '/etc/bash.bashrc'
        for var, value in self.config.get('env_vars', {}).items():
            var_string = 'export {var}={value}'.format(var=var.upper(), value=value)
            env_vars.append(var_string)
        vars_string = "bash -c \"echo -e '{env_vars}' | tee -a {rcfile}\"".format(
            env_vars='\n'.join(env_vars), rcfile=rcfile)
        self.exec_cmd(vars_string)

    def restart_instance(self):
        res = {}
        logger.info('Restarting instance: %s %s',
                    self.__full_config.instance_config.get('task_id'),
                    self.__full_config.instance_config.get('customer_id'))
        retries = 0
        # Restart instance and check if it's running
        try:
            # Check if supervisor is running
            supervisor_status = self.check_supervisor()
            if not supervisor_status:
                res.update({'error': ('Failed to restart the instance: Supervisor'
                                      ' process could not start')})
                return res
            self.exec_cmd('supervisorctl stop ' + self.supervisor_group)
            # Check if odoo was successfully stopped
            while retries <= 3:
                python_process = self.exec_cmd('pgrep python -a').split('\n')
                odoo_process = [process for process in python_process if 'odoo' in process]
                if not odoo_process:
                    break
                logger.info('waiting for the odoo process to stop')
                retries += 1
                time.sleep(20)
            if self.is_running or odoo_process:
                state = self.exec_cmd('supervisorctl status odoo')
                res.update({
                    'error': 'Failed to restart the instance, could not stop odoo'})
                return res
            self.exec_cmd('supervisorctl start ' + self.supervisor_group)
            if not self.is_running:
                state = self.exec_cmd('supervisorctl status ' + self.supervisor_app)
                res.update({
                    'error': ('Failed to restart the instance, could not'
                              ' start odoo: {state}').format(state=state.strip())})
                return res
        except (errors.SupervisorStatusError, errors.CommandError, errors.NotRunning) as error:
            res.update({'error': 'Couldn\'t restart the instance, {error}'.format(
                error=utils.get_error_message(error))})
            return res
        res.update({'result': 'Instance successfully restarted'})
        logger.info('Done')
        return res

    def restart_container(self):
        # Get the information of the container
        pg_shell = PostgresShell(utils.odoo2postgres(self.db_config))
        try:
            container_info = self.basic_info
        except errors.NoSuchContainer:
            logger.error('Failed to restart the container, it does not exist')
            return False
        # Restart container
        try:
            logger.info('Stopping container %s', container_info.get('name'))
            container_obj = self.cli2.containers.get(container_info.get('Id'))
            container_obj.stop()
            logger.info('Terminating connections on the database %s', self.db_config['db_name'])
            pg_shell.terminate_sessions(
                self.db_config['db_name'], self.db_config.get('db_owner'),
                self.db_config.get('db_owner_passwd')
            )
            logger.info('Starting container %s', container_info.get('name'))
            container_obj.start()
        except APIError:
            logger.error('Failed to restart the container, it took to long to start/stop')
            return False
        # Get the container object again so it gets the updated state and
        # check if it is running
        container_obj = self.cli2.containers.get(container_info.get('Id'))
        retries = 0
        while retries <= 3:
            if container_obj.status == 'running':
                return True
            time.sleep(5)
            retries += 1
        # If the while loop ends, that means the container failed to start
        logger.error('Failed to restart the container, state: %s', container_obj.status)
        return False

    def extras_deactivate(self):
        """Method find the files deactivate.sql that is into of the repositories
        of the container and return a dict with the queries
        """
        res = {}
        env_vars = self.config.get('env_vars') or self.docker_env
        odoo_home = env_vars.get('odoo_home')
        cmd = "find {0}/instance -name 'deactivate.jinja'".format(odoo_home)
        wf_path = self.__full_config.working_folder
        name = path.basename(wf_path)
        try:
            res_cmd = self.exec_cmd_image(cmd, {name: '/tmp'}, wf_path)
        except errors.CommandError:
            return res
        if not res_cmd.get('result') or 'No such file or directory' in res_cmd.get('result') or\
                res_cmd.get('error'):
            logger.info('Could not find the deactivate.jinja file in %s', odoo_home)
            res.update(res_cmd)
            return res
        for file_sql in res_cmd['result'].splitlines():
            if not file_sql:
                continue
            sqls = self.exec_cmd_image("cat {0}".format(file_sql), {name: '/tmp'}, wf_path)
            if sqls.get('error'):
                logger.error(sqls.get('error'))
                res.update(sqls)
                return res
            sql_json = json_helper.load_json(Template(sqls['result']).render(**env_vars))
            res.update(sql_json or {})
        return res

    def start_maintenance_page_container(self):
        """Starts the container that will be used while the production
        instance is updated. This container will be created with an image
        that will serve a custom page or app, and will be destroyed after the
        update is done.
        """
        logger.info('Starting maintenance page\'s container')
        ports = {'80': self.config.get('ports').get('8069')}
        container_config = {
            "name": (
                '%(customer_id)s_maintenance_page' % {
                    'customer_id': self.__full_config.instance_config.get('customer_id')
                }
            ),
            "image": self.__full_config.maintenance_page.get("image_name"),
            "hostname": container.generate_hostname(self.__full_config),
            "ports": container.generate_port_lists(ports),
            "environment": container.generate_env_vars(
                self.__full_config.maintenance_page.get("env_vars")
            ),
        }
        host_config = {
            "port_bindings": container.generate_port_bindings(ports),
            "mem_limit": self.config.get("mem_limit", 0),
            "memswap_limit": self.config.get("mem_limit", 0),
            "restart_policy": {
                "MaximumRetryCount": 0,
                "Name": "unless-stopped"
            }
        }
        hconfig = self.cli.create_host_config(**host_config)
        container_obj = self.cli.create_container(
            host_config=hconfig, **container_config
        )
        container_id = container_obj.get("Id")
        logger.info('Container created %s', container_id)
        logger.info('Starting container %s', container_id)
        self.cli.start(container=container_id)
        logger.info('Container started %s', container_id)
        return container_id

    def create_vx_schema(self, db_config):
        """Create vx schema and required functions to anonymize data
        :return: boolean, error_message (True when process is successful)"""
        try:
            self.execute_query(
                db_config, "CREATE SCHEMA vx; SECURITY LABEL FOR anon ON SCHEMA vx IS 'TRUSTED';")
        except Exception as e:
            error = utils.get_error_message(e)
            logger.error('Error while create vx schema, and set security label: %s', error)
            return False, error
        queries = pg_anon.get_queries_vx_functions()
        try:
            self.execute_query(db_config, queries)
        except Exception as e:
            error = utils.get_error_message(e)
            logger.error('Error while create functions in vx schema: %s', error)
            return False, error
        return True, ''

    def anonymize_db(self, db_config):
        """Apply masked functions to anonymize some data, as indicated by config file.
        Verify that postgresql anonymizer is installed, otherwise, try to install it.

        :param db_config: Configuration for database where the query will be executed
        :return: boolean, error_message (True when process is successful)"""
        # verify if postgresql anonymizer extension is available
        res = self.execute_query(
            db_config, "SELECT name, installed_version FROM pg_catalog.pg_available_extensions "
                       "WHERE name='anon'")
        if res:
            if res[0]['installed_version'] is None:
                query = "ALTER DATABASE {} SET session_preload_libraries = 'anon';".format(
                    db_config['dbname'])
                try:
                    self.execute_query(db_config, query)
                except Exception as e:
                    error = utils.get_error_message(e)
                    logger.error('Error while set anon extension: %s', error)
                    return False, 'Anonymizer error: {}'.format(error)
                query = "CREATE EXTENSION anon CASCADE; SELECT anon.init(); " \
                        "ALTER DATABASE {} SET anon.restrict_to_trusted_schemas = on;" \
                    .format(db_config['dbname'])
                try:
                    self.execute_query(db_config, query)
                except Exception as e:
                    error = utils.get_error_message(e)
                    logger.error('Error while create anon extension: %s', error)
                    return False, 'Anonymizer error: {}'.format(error)

            res = self.execute_query(db_config,
                                     "SELECT COUNT(*) as qty FROM pg_catalog.pg_namespace "
                                     "WHERE nspname='vx';")
            if not res[0]['qty']:
                res, error = self.create_vx_schema(db_config)
                if not res:
                    return res, 'Anonymizer error: {}'.format(error)

            queries = pg_anon.get_queries_from_operations_file(
                '/etc/deployv/deactivate/anon_operations.csv')
            for item in queries:
                try:
                    self.execute_query(db_config, item)
                except Exception as e:
                    error = utils.get_error_message(e)
                    logger.error('Error while executing anonymize data query: %s', error)
                    return False, 'Anonymizer error: {}'.format(error)
            return True, ''
        logger.error('Extension postgresql anonymizer is not available')
        return False, 'Anonymizer error: Extension postgresql anonymizer is not available'

    def deactivate_database(self, db_name, extra_config=False):
        """Executes the deactivation queries found in the deactivation.jinja file
        on the provided database.

        :param db_name: Name of the database where the queries will be executed
        :param extra_config: Dictionary with the additional config used to
            render the deactivation.jinja file

        :return: Error message when data anonymization fails, None otherwise
        """
        logger.info('Deactivating database %s', db_name)
        db_config_user = self.db_config.copy()
        db_config_user.update({'db_name': db_name})
        db_config_owner = db_config_user.copy()
        db_config_owner.update({
            'db_user': self.db_config.get("db_owner"),
            'db_password': self.db_config.get('db_owner_passwd')
        })
        db_config_user = utils.odoo2postgres(db_config_user)
        db_config_owner = utils.odoo2postgres(db_config_owner)
        # executes anonymization, if required
        if path.isfile('/etc/deployv/deactivate/anon_operations.csv'):
            logger.info('Start anonymization of data')
            res, error_msg = self.anonymize_db(db_config_user)
            if res:
                logger.info('Successful anonymization of data')
            else:
                logger.exception('Failed anonymization of data')
                postgres = PostgresShell(utils.odoo2postgres(self.db_config.copy()))
                postgres.drop(db_name)
                return {'error': error_msg}
        extra_config = extra_config or {}
        extra_sqls = self.extras_deactivate()
        if extra_sqls.get('error'):
            logger.error(extra_sqls.get('error'))
            return extra_sqls
        query_config = ("select id from ir_model_fields where name"
                        "= 'value' and model = 'ir.config_parameter';")
        config_parameter_id = self.execute_query(db_config_user, query_config)
        config_parameter_id = config_parameter_id and config_parameter_id[0].get('id')
        extra_config.update({
            'field_id': config_parameter_id, 'uuid': uuid4(),
            'migration': self.__full_config.migration})
        # Sqls to be executed
        template_path = deployv_static.get_template_path('deactivation.jinja')
        with open(template_path, 'r') as obj:
            json_template = obj.read()
        sqls = json_helper.load_json(Template(json_template).render(extra_config))
        # Load the demo pac
        certificate_pac = utils.get_certificates_pac()
        sql = "UPDATE l10n_mx_edi_certificate SET "
        sql += " ,".join(["%s='%s'" % (k, v) for k, v in certificate_pac.items()])
        sqls.update({"mx_edi_certificate": sql})
        sqls.update(extra_sqls)
        # Start to execute the queries
        try:
            for name in sorted(sqls):
                logger.debug('Running %s', name)
                sql = sqls[name]
                try:
                    self.execute_query(db_config_user, sql, allow_errors=True)
                except psycopg2.errors.InsufficientPrivilege as error:
                    logger.warning("Inssufficient permissions for user '%s': %s",
                                   db_config_user.get('user'), utils.get_error_message(error))
                    self.execute_query(db_config_owner, sql, allow_errors=True)
        except psycopg2.OperationalError as error:
            logger.exception('Could not deactivate the database: %s', error.message)
        return {}

    @property
    def db_config(self):
        """ Parse env vars from a container to get the needed parameters to dump the database

        :return: dict with the needed configuration parameter
        """

        env_vars = self.docker_env or {}
        if self.__full_config.backup_type in ['btrfs', 'zfs']:
            env_vars = env_vars or self.config.get('env_vars')
        res = self.__full_config.instance_config.get("config").copy()
        logger.debug(json.dumps(env_vars, sort_keys=True, indent=4))
        logger.debug("Instance config dict from json: %s", str(res))
        res.update({
            'db_host': res.get('db_host') or env_vars.get('db_host') or '127.0.0.1',
            'db_port': env_vars.get('external_db_port') or res.get(
                'db_port') or env_vars.get('db_port') or 5432,
            'db_user': res.get('db_user') or env_vars.get('db_user'),
            'db_password': res.get('db_password') or env_vars.get('db_password'),
            'data_dir': self.get_data_dir()
        })
        docker_network_info = self.cli.inspect_network('bridge')
        docker_gateway = docker_network_info.get('IPAM').get('Config')[0].get('Gateway')
        is_docker_ip = utils.is_docker_ip(res.get('db_host'), docker_gateway)
        if ((is_docker_ip or res.get('db_host') == 'localhost') and not
                self.__full_config.deployer.get('deployv_in_docker')):
            res.update({'db_host': '127.0.0.1'})
        return res

    def get_data_dir(self):
        """Return the path where store the filestore inside of container.
        """
        try:
            inspected = self.inspect()
        except errors.NoSuchContainer:
            return ''
        for volume in inspected.get('Mounts', []):
            if volume and '.local/share/Odoo' in volume.get('Destination'):
                return volume.get('Source')
        logger.error('Datadir not found, wont be able to backup attachments')

    def save_repositories(self, filename=False):
        """Save the repositories stored in the image name.
        that information is saved in the specify path.

        :param file_name: The file's name where the repositories's information are saved.
        :type: str
        """
        if not self.temp_folder:
            logger.warning("Save Repositories: Doesn't exist a temp folder where save the file")
            return ''
        home = self.__full_config.container_config.get('env_vars').get('odoo_home')
        odoo_user = self.__full_config.container_config.get('env_vars').get('odoo_user')
        filename = filename or 'branches.json'
        self.exec_cmd('bash -c "rm -rf /tmp/%s"' % (filename))
        cmd = 'bash -c "branchesv save -p %s/instance -f /tmp/%s"' % (home, filename)
        file_path = path.join(self.temp_folder, filename)
        cmd_output = self.exec_cmd(cmd)
        logger.debug('Branchesv result:\n%s', cmd_output)
        if not path.exists(file_path):
            cmd = 'su %s -c "branchesv save -p %s/instance -f /tmp/%s"' % (
                odoo_user, home, filename)
            cmd_output = self.exec_cmd(cmd)
            logger.debug('Branchesv result:\n%s', cmd_output)
        return file_path

    @property
    def config_variables(self):
        """Property to get the value of variables from this Class or
        get the information from the docker environment.
        Then if the `BASE_IMAGE`, `PSQL_VERSION`, `TRAVIS_PYTHON_VERSION`, `ODOO_REPO` or
        `ODOO_BRANCH` keys doesn't exist, it will set default values according to the version.
        """
        docker_env = self.docker_env or {}
        variables = self.__full_config.variables or docker_env
        if self.__full_config.container_config.get('env_vars'):
            variables.update(self.__full_config.container_config.get('env_vars'))
        main_repo = [r for r in self.__full_config.instance_config.get('repositories')
                     if r.get('main_repo')]
        if docker_env.get('main_app'):
            variables.update({'main_app': docker_env.get('main_app')})
        if not variables.get('version') and main_repo:
            variables.update({'version': main_repo[0].get('branch')})
        version = variables.get('version') or ''
        default_variables = {
            'base_image': 'vauxoo/odoo-%s-image' % (version.replace('.', '')),
            'psql_version': PSQL_VERSIONS.get(version),
            'travis_python_version': TRAVIS_VERSIONS.get(version),
            'odoo_repo': 'vauxoo/odoo', 'odoo_branch': version
        }
        for key, value in default_variables.items():
            if not variables.get(key) and value:
                variables.update({key: value})
        return variables

    def _get_db_manifest(self, tmp_dir):
        """ Generates a manifest with the information of the modules installed in the database
        to be added in the backups, just like the one added by odoo in its backups.
        """
        db_config = self.db_config
        with PostgresConnector(utils.odoo2postgres(db_config)) as db:
            try:
                installed_modules = db.execute(
                    "SELECT name, latest_version FROM ir_module_module WHERE state = 'installed'",
                    raw=True
                )
                version = db.execute("SELECT current_setting('server_version_num');")
            except psycopg2.ProgrammingError as error:
                error_msg = utils.get_error_message(error)
                logger.error('An error ocurred while obtaining the dump manifest: %s', error_msg)
                raise errors.CommandError('backup', error_msg)
        version = int(version[0].get('current_setting'))
        manifest = {
            'odoo_dump': 1,
            'db_name': db_config.get('db_name'),
            'pg_version': "%d.%d" % divmod(version / 100, 100),
            'modules': dict(installed_modules)
        }

        # Instances that use python3 have python2 as default, that version does not have
        # odoo's requirements so it raises an error when we try to execute this script,
        # so we need to check which python version to use. It's done by checking if
        # python3 exists instead of checking which version of odoo is running to avoid
        # having to update the list with each new version of odoo.
        python3_exists = self.exec_cmd('ls /usr/bin/python3')
        pybin = 'python' if 'No such file or directory' in python3_exists else 'python3'
        cmd = '{pybin} /tmp/get_release_data.py'.format(pybin=pybin)
        release_data = self.exec_cmd(cmd)

        # Backwards compatibility condition, only the new images will have this script
        # so we will create it manually if the container does not have it yet. This
        # block needs to be removed in the future.
        if 'No such file or directory' in release_data:
            script = """
import sys

sys.path.insert(1, '/home/odoo/instance/odoo')
from odoo import release
import simplejson as json

print(json.dumps({
    'version': release.version,
    'major_version': release.major_version,
    'version_info': release.version_info}))
            """
            with open(path.join(self.temp_folder, 'get_release_data.py'), 'w') as obj:
                obj.write(script)
            release_data = self.exec_cmd(cmd)

        # Parse release data obtained from the script
        parsed_data = json_helper.load_json(release_data)
        # When this method is called by an external tool (e.g. a script), the parsed data
        # cannot be obtained because it may not send the complete container information,
        # returning False in the method above and raising an error when we try to update
        # the response.
        manifest.update(parsed_data or {})
        manifest_path = path.join(tmp_dir, 'manifest.json')
        json_helper.save_json(manifest, manifest_path)
        return manifest_path

    def create_extension_unaccent(self, db_name):
        """ Logs into postgres as a superuser in order to create the
        unaccent extension.

        :param db_name: name of the database where the extension will
            be created.
        """
        pg_shell = PostgresShell(self._get_admin_config())
        pg_shell.create_extension_unaccent(db_name)

    def _get_admin_config(self):
        admin_config = self.db_config.copy()
        admin_config.update({
            'db_user': self.__full_config.postgres.get('db_user'),
            'db_password': self.__full_config.postgres.get('db_password'),
        })
        return utils.odoo2postgres(admin_config)


class DummyInstanceV:

    def __init__(self, config, timeout=False, docker_url=False):
        super(DummyInstanceV, self).__init__()



# coding: utf-8

"""DockerV - Docker helper class by Vauxoo
==========================================

This module provides a helper class to deal with docker containers in an OO way, basically wraps
`docker-py <https://github.com/docker/docker-py>`_ to deal with a json formatted configuration.

Most of the parameters can be checked in the docker official `documentation
<https://docs.docker.com/reference/run/>`_ or in the `docker-py api documentation
<http://docker-py.readthedocs.org/en/latest/api/>`_.

Still not all functionalities are implemented, but the minimal to start a docker container with
basic configuration::

    {
        "apt_install": {},
        "pip_install": {},
        "domain": "localhost",
        "command": "sleep 20",
        "env_vars": {
            "odoo_config_file": "/home/odoo/.openerp_serverrc",
            "odoo_home": "/home/odoo",
            "odoo_user": "odoo"
        },
        "image_name": "busybox",
        "mem_limit": "768m",
        "ports": {
            "8069": 8069,
            "8072": None
        },
        "remove_previous": True,
        "working_folder": "/path/to/docker_volumes",
        "volumes":{
            "filestore": "/home/odoo/.local/share/Odoo",
            "logs": "/var/log/supervisor",
            "ssh": "/home/odoo/.ssh",
            "tmp": "/tmp"
        }
    }

Where:

* *apt_install* (not implemented yet): Install additional packages using apt-get install.
    (This will be changed to use a generic way of installing packages in other distros)
* *pip_install* (not implemented yet): Install additional dependencies using pip.
* *domain*: Domain where the container will be executing.
* *image_name*: Docker image name that will be used to create the container.
    (See `official documentation <https://docs.docker.com/userguide/dockerimages/>`_)
* *mem_limit*: Max amount of ram allowed for the container.
* *ports*: A dict with ports mapping, key is the port number inside the container, value the host
    port.  If None is provided as value the port number will be assigned randomly by docker
    service.
* *remove_previous*: If there is a container with the same name or id it will be removed before
    launching the new one.
* *working_folder*: Folder in the host where the shared volumes (if any) will be created.
* *volumes*: A dict with the volume mapping, keys are the folder names inside working_folder and
    values are the actual path inside the container.
"""

import logging
from os import path
from docker import DockerClient
from docker import APIClient as Client
import docker.errors
import simplejson as json
from deployv.helpers import container
from deployv.helpers.utils import get_error_message, decode
from deployv.base import errors
from retry_decorator import retry
from time import sleep


logger = logging.getLogger(__name__)  # pylint: disable=C0103


class DockerV:
    """Helper class to make docker container deployments easier by code.

    If your docker service is not using the default socket you can change it by passing
    docker_url::

        config = {
            'image_name': 'busybox',
            'command': 'sleep 3',
            'mem_limit': '16m',
            'container_name': 'test_deployv',
            'ports' : {
                '8069': 8069,
                '8072': None
            },
            'env_vars': {
                'var1':'value1',
                'var2': 1234,
            }
        }
        container_object = DockerV(config)
        container_id = container_object.deploy_container()

    ``image_name`` and ``command`` are the only required parameters, others parameters are
    optional. If a port is empty, it will be autoasigned by docker.

    If the container already exists the config dict would be something like::

        config = {
            'container_name': 'test_deployv',
            'id': 'container_id'
        }

    Only one of them is needed. You can specify both, but id will be checked first then the
    container_name.
    """

    __config = {}
    __docker_id = None
    __url = None

    def __init__(self, config, timeout=300, docker_url="unix://var/run/docker.sock"):
        self.__cli = Client(base_url=docker_url, timeout=timeout)
        self.__cli2 = DockerClient(base_url=docker_url, timeout=timeout)
        self.__config = config
        self.__url = docker_url
        if self.__config.get('id', False):
            inspected = self.inspect(self.__config.get('id'))
            self.__docker_id = inspected.get('Id')
        elif self.__config.get('container_name', False):
            try:
                inspected = self.inspect(self.__config.get('container_name'))
            except errors.NoSuchContainer as error:
                error_message = get_error_message(error)
                logger.debug('%s', error_message)
            else:
                self.__docker_id = inspected.get('Id')
                logger.debug('Container inspected %s',
                             json.dumps(inspected, sort_keys=True, indent=2))
                self.__config.update({'id': inspected.get('Id')})
                if not self.__config.get('image_name'):
                    self.__config.update({
                        'image_name': inspected.get('Config').get('Image')})
                if not self.__config.get('env_vars'):
                    env_vars = container.parse_env_vars(inspected.get('Config').get('Env'))
                    self.__config.update({'env_vars': env_vars})

    @property
    def url(self):
        return self.__url

    @property
    def cli(self):
        """This is the low-level API (backwards compatible with docker 1.X).

        Read the docs in: https://docker-py.readthedocs.io/en/stable/api.html
        """
        return self.__cli

    @property
    def cli2(self):
        """This is the high-level API (introduced in docker 2.0).

        Read the docs in: https://docker-py.readthedocs.io/en/stable/client.html
        """
        return self.__cli2

    @property
    def config(self):
        return self.__config

    @config.setter
    def config(self, value):
        keys = [u'image_name', u'command', ]
        if not all([key in value for key in keys]):
            raise KeyError('Mandatory keys: {}'.format(', '.join(keys)))
        self.__config = value

    @property
    def docker_id(self):
        """Returns container id, the same you see when executing 'docker ps' in the first column.

        :returns: Container Id.
        :rtype: str
        """
        return self.__docker_id.lower() if self.__docker_id else None

    @property
    def docker_env(self):
        """Gets env vars from a docker container.

        :returns: dict with var name as key and value.
        :rtype: dict
        """
        try:
            inspected = self.inspect()
        except errors.NoSuchContainer as error:
            logger.debug(get_error_message(error))
            return {}

        env_vars = inspected.get('Config').get('Env')
        res = container.parse_env_vars(env_vars)
        return res

    def deploy_container(self):
        """Deploys a container with the given config and returns some basic information about the
        container. See :func:`~deployv.base.dockerv.DockerV.basic_info` for more information about
        the format and content.

        :returns: Some basic information about the container.
        :rtype: dict
        """
        env_vars = container.generate_env_vars(self.config.get("env_vars"))
        container_name = self.config.get("container_name").lower()
        logger.debug('Deploying container %s', container_name)
        working_folder = path.join(
            self.__config.get("working_folder"),
            container_name)
        volume_binds = container.generate_binds(
            self.__config.get("volumes"), working_folder)
        volumes = list(self.__config.get("volumes").values())
        if self.config.get('full_stack') and self.config.get('pg_log_postgres'):
            volume_binds.update({self.config.get('pg_log_postgres'): {
                'bind': '/var/log/pg_log', 'ro': False}})
            volumes.append('/var/log/pg_log')
        self.set_default_ports()
        try:
            self.create_container(env_vars, volume_binds, volumes)
        except errors.ErrorPort:
            if self.instance_type != 'develop':
                raise
            self.config['ports'] = {
                port: 0 for port in self.config['ports']
            }
            self.set_default_ports()
            self.create_container(env_vars, volume_binds, volumes)
        return self.basic_info

    def remove_container(self, container_name=False):
        """Stops and removes the specified container. Docker id or name must be in lowercase, if
        not, it will be converted to lowercase.
        """
        if not container_name:
            container_name = self.docker_id
        try:
            self.__cli.stop(container_name)
            self.__cli.remove_container(container_name, force=True)
        except docker.errors.NotFound as error:
            logger.debug(get_error_message(error))
        else:
            self.__docker_id = None

    def remove_image(self, image_name=False):
        """Try remove the specified image.
        returns a list with message if it succeeded in removing the image,
        else return a list empty

        :param image_name: image name to remove
        :return: list
        """
        res = []
        image_id = self.__cli.images(name=image_name, quiet=True)
        image_id = image_id and image_id[0]
        if not image_id:
            return res
        try:
            self.__cli.remove_image(image_name)
            res.append("Removed image {}".format(image_name))
        except docker.errors.APIError:
            pass
        return res

    @property
    def basic_info(self):
        """Gets some basic information form the container::

            {
                'Id': 'container id',
                'name': 'container name',
                'image_name': 'image used to build the container',
                'status': 'container status (same shown when "docker ps" is executed)',
                'ports': 'dict with port assignments {local (inside docker) : public}'
            }

        :returns: Container information.
        :rtype: dict
        """
        res = {}
        try:
            inspected = self.inspect()
        except errors.NoSuchContainer:
            return res
        if inspected.get('Id') == self.docker_id:
            res.update(
                {
                    'Id': inspected.get('Id'),
                    'name': inspected.get('Name')[1:],
                    'image_name': inspected.get('Config').get('Image'),
                    'status': 'running' if inspected.get('State') else 'exited',
                    'ports': container.get_ports_dict(inspected)
                })
        return res

    def get_container_packages(self):
        """Returns the installed packages inside the container (for pip and apt).

        The packages are returned in a dict with the package manager and the installed packages
        with their versions inside::

            {
                'pip': {
                    'package': 'version',
                    ...
                },
                'apt': {
                    'package': 'version',
                    ...
                },
            }

        :returns: Installed packages
        :rtype: dict
        """
        # pylint: disable=R1717
        pip_res = self.exec_cmd(['pip', 'freeze'])
        pip = dict([line.split('==') for line in pip_res.strip().split('\n') if '==' in line])
        apt_res = self.exec_cmd(['apt', 'list', '--installed'])
        apt = dict([(pck.split('/')[0] for pck in line.split(' ')[:2])
                    for line in apt_res.split('\n') if 'installed' in line])
        return {
            'pip': pip,
            'apt': apt,
        }

    @retry(errors.ClientAPIError, tries=3, timeout_secs=0.5, logger=logger)
    def exec_cmd(self, cmd, user=None):
        """Wraps and logs the `exec_create
        <https://docker-py.readthedocs.org/en/latest/api/#exec_create>`_ and `exec_start
        <https://docker-py.readthedocs.org/en/latest/api/#exec_start>`_ methods.
        :param cmd: Command line to be executed inside the Docker container.
        :type cmd: str
        :returns: The command output.
        """
        logger.debug('exec_cmd: %s', cmd)
        try:
            exec_id = self.__cli.exec_create(self.docker_id, cmd, user=user)
        except (docker.errors.APIError, docker.errors.NullResource) as error:
            error_msg = get_error_message(error)
            if 'is not running' in error_msg:
                raise errors.NotRunning(self.docker_id)
            if 'Resource ID was not provided' in error_msg:
                raise errors.CommandError(cmd, "Container does not exist")
            if hasattr(error, 'is_client_error') and error.is_client_error():
                raise errors.ClientAPIError(
                    "Container {0} is in conflict or stopped unexpectedly".format(self.docker_id))
            raise
        res = decode(self.__cli.exec_start(exec_id.get('Id')))
        return res

    def inspect(self, container_id=False):
        """Wrapper for docker-py.inspect _container.

        :param container_id: Optional parameter with the container name or id.
        :type container_id: str
        :returns: Information about the container.  See `inspect_container in the official docs
            <https://docker-py.readthedocs.org/en/latest/api/#inspect_container>`_.
        """
        container_id = container_id or self.docker_id
        try:
            res = self.cli.inspect_container(container_id)
        except docker.errors.NotFound:
            raise errors.NoSuchContainer(container_id)
        except docker.errors.NullResource:
            raise errors.NoSuchContainer(container_id)
        return res

    def install_packages(self, apt_packages=None, pip_packages=None):
        """Installs apt and/or pip packages inside the container and returns the full operation
        log. Runs quietly, so the log is cleaner.

        The result has the following format (if no errors were found)::

            {
                'apt_install': [
                    {
                        'package_name': {
                            'installed': True
                        }
                    }
                ]
                'pip_install': [
                    {
                        'package_name': {
                            'installed': True
                        }
                    }
                ]
            }

        In case of error::

            {
                'apt_install': [
                    {
                        'package_name': {
                            'installed': False,
                            'message': 'Error msg generated by the installer'
                        }
                    }
                ]
                'pip_install': [
                    {
                        'package_name': {
                            'installed': False,
                            'message': 'Error msg generated by the installer'
                        }
                    }
                ]
            }

        The results can be mixed if some packages fail and some don't during the installation
        process.

        :param apt_packages: Packages to be installed via apt.
        :type apt_packages: list
        :param pip_packages: Packages to be installed via pip.
        :type pip_packages: list
        :returns: Full log operations.
        :rtype: dict
        """
        res = {}
        if isinstance(apt_packages, list) and apt_packages:
            self.exec_cmd('apt-get update -q')
            res.update({'apt_install': []})
            for apt_package in apt_packages:
                cmd_res = self.exec_cmd('apt-get install {name} -yq'.format(name=apt_package))
                logger.debug('apt_install %s', cmd_res)
                composed = self._compose_message('apt_install', apt_package, cmd_res)
                res.get('apt_install').append(composed)

        if isinstance(pip_packages, list) and pip_packages:
            res.update({'pip_install': []})
            for pip_package in pip_packages:
                cmd_res = self.exec_cmd('pip install {name}'.format(name=pip_package))
                logger.debug('pip_install %s', cmd_res)
                composed = self._compose_message('pip_install', pip_package, cmd_res)
                res.get('pip_install').append(composed)
        return res

    def _compose_message(self, package_installer, package, result):
        """Composes the response.

        The response has the following format::

            {
                'package_name': {
                    'installed': Boolean,
                    'message': 'if any error, this will have the error msg.'
                               'if no error, this won't be present.'
                }
            }

        :param package_installer: Package installer name (so far pip or apt).
        :type package_installer: str
        :param package: Package/dependency installed.
        :type package: string.
        :param result: The result from :meth:`~deployv.base.dockerv.DockerV.exec_cmd`.
        :returns: Formatted output in a json format.
        :rtype: dict
        """
        res = {package: {}}
        errors_txt = ['Unable to locate package', 'No matching distribution found']
        if any([error_txt in result for error_txt in errors_txt]):
            res.get(package).update(
                {
                    'installed': False,
                    'message': result
                })
        else:
            res.get(package).update({'installed': True})
        return res

    @retry(docker.errors.APIError, tries=3, timeout_secs=0.5, logger=logger)
    def pull(self, image_name=False, raise_exception=False):
        """Make a pull of image with provided name.

        :param image_name: Name of image to pull.
        :type image_name: str
        :param raise_exception: Indicates if raises an exception when no image got from cli's pull
        :type raise_exception: bool
        """
        image_name = image_name or self.config.get('image_name')
        try:
            self.cli_pull_wrapper(image_name=image_name)
        except errors.NoSuchImage as exc:
            if raise_exception:
                raise exc
            image = self.inspect_image(image_name)
            logger.debug('Image info %s', image)

    def cli_pull_wrapper(self, image_name=False):
        """This is a simple wrapper for docker pull, to use the deployv exceptions instead of the
        docker ones.

        :param image_name: The image name to pull.
        :type image_name: str
        :raises: :class:`~deployv.base.errors.NoSuchImage` if the image doesn't exist.
        """
        try:
            self.cli.pull(image_name)
        except (docker.errors.NotFound, docker.errors.APIError) as error:
            msg_error = get_error_message(error)
            if (isinstance(error, docker.errors.NotFound) or
                    'was deleted or has expired' in msg_error):
                logger.debug('Image not found in hub')
                error = errors.NoSuchImage(image_name)
            raise error

    @retry(docker.errors.APIError, tries=3, timeout_secs=0.5, logger=logger)
    def push(self, image_name):
        """Pushes an image and logs (in DEBUG level) the process.

        :param image_name: The image that will be pushed.
        :type image_name: str
        """
        logger.info('Pushing image "%s"', image_name)
        res = {'result': {'log': []}}
        for line in self.cli.push(image_name, stream=True):
            obj = json.loads(decode(line))
            res['result']['log'].append(obj)
            logger.debug(obj)
            if obj.get('error'):
                return {
                    'error': obj.get('error').strip(),
                }
            if obj.get('aux'):
                res['result'].update({'details': obj.get('aux')})
        return res

    def commit(self, repository, tag):
        """Commits the container with the given repository and tag.

        :param repository: The repository for the image.
        :type repository: str
        :param tag: The tag for the image.
        :type tag: str
        """
        res = {'result': {}}
        try:
            container_info = self.basic_info
        except errors.NoSuchContainer:
            res.update({'error': 'Specified container doesn\'t exist'})
            return res
        logger.info('Committing container "%s" as "%s:%s"',
                    container_info.get('name'), repository, tag)
        container_id = container_info.get('Id')
        try:
            commit = self.cli.commit(container_id, repository=repository, tag=tag)
        except docker.errors.APIError as error:
            res.update({'error': str(error)})
            return res
        image = self.cli.inspect_image(commit.get('Id'))
        image_tagged = image.get('RepoTags')[0]
        res['result'].update({
            'container_info': container_info,
            'tagged_image': image_tagged,
        })
        return res

    def commit_and_push(self, repository, tag):
        """Commits an image with the given repository and tag, and pushes it.

        :param repository: The repository for the image.
        :type repository: str
        :param tag: The tag for the image.
        :type tag: str
        """
        res = {'result': {}}
        commit = self.commit(repository, tag)
        res['result'].update(commit.get('result'))
        if commit.get('error'):
            res.update({'error': commit.get('error')})
            return res
        push = self.push(commit.get('result').get('tagged_image'))
        res['result'].update({'push_result': push.get('result')})
        if push.get('error'):
            res.update({'error': push.get('error')})
            return res
        return res

    def tag(self, image, repository, tag):
        """Tags an already created image with a new repository and tag.

        :param image: The image that will be tagged.
        :type image: str
        :param repository: The new repository for the image.
        :type repository: str
        :param tag: The new tag for the image.
        :type tag: str
        """
        res = {'result': {}}
        tagged_image = '{repo}:{tag}'.format(repo=repository, tag=tag)
        logger.info('Tagging image "%s" as "%s"', image, tagged_image)
        try:
            tag = self.cli.tag(image, repository, tag=tag)
        except docker.errors.APIError as error:
            res.update({'error': str(error)})
            return res
        res['result'].update({
            'tagged_image': tagged_image
        })
        return res

    def tag_and_push(self, image, repository, tag):
        """Tags an already created image with a new repository and tag, and pushes the new image.

        :param image: The image that will be tagged.
        :type image: str
        :param repository: The new repository for the image.
        :type repository: str
        :param tag: The new tag for the image.
        :type tag: str
        """
        res = {'result': {}}
        tag_result = self.tag(image, repository, tag=tag)
        res['result'].update(tag_result.get('result'))
        if tag_result.get('error'):
            res.update({'error': tag_result.get('error')})
            return res
        push = self.push(tag_result.get('result').get('tagged_image'))
        res['result'].update({'push_result': push.get('result')})
        if push.get('error'):
            res.update({'error': push.get('error')})
            return res
        return res

    def inspect_image(self, image_name):
        """Inspect image with provided name, returning its params.

        :param image_name: The image name that will be inspected.
        :type image_name: str
        :returns: The result of docker inspect_image.
        :raises: :class:`~deployv.base.errors.NoSuchImage` when image doesn't exist
        """
        try:
            self.cli_pull_wrapper(image_name)   # just in case does not exist in local
        except errors.NoSuchImage:
            pass
        res = self.cli_inspect_image_wrapper(image_name)
        return res

    def cli_inspect_image_wrapper(self, image_name):
        """Wrapper for the inspect image, basically does the same but raises a deployv exception
        instead of the docker-py one.

        :param image_name: The image name that will be inspected.
        :type image_name: str
        :returns: The result of docker inspect_image.
        :raises: :class:`~deployv.base.errors.NoSuchImage` if the image doesn't exist.
        """
        try:
            res = self.cli.inspect_image(image_name)
        except docker.errors.NotFound:
            raise errors.NoSuchImage(image_name)
        return res

    def set_default_ports(self):
        """Sets default port mappings if the ones specified in the json
        contain falsy values so we can have control over the ports assigned
        to the instances. It also checks if the ports specified for a
        develop instance are free, it sets new ports if they are not.
        """
        deploy_cfg = self._InstanceV__full_config.deployer
        instance_ports = self.config.get('ports')
        specified_ports = set(self._parse_ports(instance_ports))
        used_ports = set(self._get_used_ports())
        port_start = int(deploy_cfg.get('docker_start_port', 30000))
        port_end = int(deploy_cfg.get('docker_end_port', 40000))
        available_specified_ports = list(specified_ports - used_ports)
        available_default_ports = [
            p for p in range(port_start, port_end) if
            p not in used_ports
        ]
        available_ports = available_default_ports + available_specified_ports
        if not available_ports:
            raise errors.ErrorPort('There are no available ports in the range'
                                   ' {start} - {end} to deploy this instance'
                                   .format(start=port_start, end=port_end))
        ports_to_map = [
            port for port, mapping in instance_ports.items() if not mapping
        ]
        if not ports_to_map:
            # There are no ports to map, try to use the ones specified
            return
        instance_ports.update(dict(zip(ports_to_map, available_ports)))
        return

    def create_container(self, env_vars, volume_binds, volumes):
        """Tries to create a container using the the given configuration as well
         as the configuration contained in the calss instance

        :param env_vars: List of env vars, vars names are uppercased as most of
         unixlike OS ["VAR1=value1", "VAR2=234"]
        :param volume_binds: Dict with the following format:

        .. code-block:: python

            {
                "{{ working_folder }}/host_folder_name": {
                    "bind": "/volume_inside_docker",
                    "ro": False
                    },
                "{{ working_folder }}/host_folder_name_2": {
                    "bind": "/volume_inside_docker_2",
                    "ro": False
                    }
            }

        :param volumes: List of volumes inside docker
         ["volume_inside_docker", "volume_inside_docker_2"]
         """
        ports = container.generate_port_lists(self.__config.get("ports"))
        port_bindings = container.generate_port_bindings(self.config.get('ports'))
        container_config = {
            "image": self.config.get("image_name"),
            "command": self.config.get("command"),
            "hostname": self.config.get("container_hostname"),
            "ports": ports,
            "environment": env_vars,
            "volumes": [
                volume.split(':')[1] if len(volume.split(':')) > 1 else
                volume for volume in volumes
            ]
        }
        if self.config.get("container_name", False):
            container_config.update(
                {'name': self.config.get("container_name").lower()})
        host_config = {
            "binds": volume_binds,
            "port_bindings": port_bindings,
            "mem_limit": self.config.get("mem_limit", 0),
            "memswap_limit": self.config.get("mem_limit", 0),
            "restart_policy": {
                "MaximumRetryCount": 0,
                "Name": "unless-stopped"
            }
        }
        if self.instance_type == "develop":
            host_config["cap_add"] = ["SYS_PTRACE"]
        if self.__config.get("remove_previous", False):
            self.__docker_id = self.config.get("container_name").lower()
            self.remove_container()
        hconfig = self.__cli.create_host_config(**host_config)
        if "5432/tcp" in hconfig.get('PortBindings'):
            default_network_info = self.__cli.inspect_network('bridge')
            default_gateway = default_network_info.get('IPAM').get('Config')[0].get('Gateway')
            host_port = hconfig.get('PortBindings').get("5432/tcp")[0].get('HostPort')
            hconfig.get('PortBindings').update({
                "5432/tcp": [
                    {'HostIp': '127.0.0.1', 'HostPort': host_port},
                    {'HostIp': default_gateway, 'HostPort': host_port}
                ]})
        logger.debug('Container config %s', container_config)
        try:
            container_obj = self.__cli.create_container(
                host_config=hconfig, **container_config)
            self.__docker_id = container_obj.get("Id")
            logger.debug('Created container %s', self.__docker_id)
        except docker.errors.NotFound:
            logger.exception('Image not found')
            raise errors.NoSuchImage(self.config.get("image_name"))
        except docker.errors.APIError as error:
            logger.exception('Error creating the container')
            if 'You have to remove (or rename) that container' in error.explanation:
                logger.warning('The container %s already exists, skipping create step',
                               container_config.get('name'))
        try:
            self.__cli.start(container=self.__docker_id)
        except docker.errors.APIError as error:
            error_message = get_error_message(error)
            if 'port is already allocated' in error_message \
                    or 'address already in use' in error_message:
                logger.warning('Port already allocated')
                raise errors.ErrorPort(
                    'At least one of the ports used by this container are being used by another'
                    ' process, manual intervention is required. Error message:\n{msg}'
                    .format(msg=error.explanation))

    def _get_used_ports(self):
        """Method that checks the ports of all the containers in the server
        in order to list all the ports used by instances in order to decide
        if the current container's ports should be reassigned. This only checks
        for the ports used by containers because the container's ports are
        reassgined only if the ports are used by another container, if they are
        used by another process (or zombie process), an error will be raised to
        alert the users.
        """
        containers = self.cli2.containers.list()
        used_ports = set()
        for container_obj in containers:
            container_info = self.inspect(container_obj.id)
            ports = self._parse_ports(container.get_ports_dict(container_info))
            used_ports.update(ports)
        return list(used_ports)

    def _parse_ports(self, ports):
        """ Receives a dictionary with the port mappings between the container
        and the host and parses them to return a list of the host ports used by
        the containers.

        :param ports: dictionary of port bindings
        :return: List of integers representing the host ports used by the host
        """
        parse_ports = {
            int: lambda x: x,
            str: lambda x: int(x.split(':')[-1]),
        }
        parse_ports[list] = lambda x: [
            parse_ports[type(port)](port) for port in x if
            port and isinstance(port, (int, str))
        ]
        res = set()
        for port in ports.values():
            parse_func = parse_ports.get(type(port))
            if not parse_func:
                continue
            parsed_ports = parse_func(port)
            if not parsed_ports:
                continue
            res.update(parsed_ports if
                       isinstance(parsed_ports, list) else [parsed_ports])
        return list(res)

    def exec_cmd_image(self, command, volumes, working_folder=False, image=False):
        """Execute the command in a specify image, also it binds the volumes with
        the working folder to store the information that doesn't want to delete

        :param command: The command that will execute in the image.
        :type: str
        :param working_folder: folder in the host where the volumes mounted
            inside the container will be created.
        :type: str
        :param volumes: The volumes that will bind.
        :type: dict with the format: {"host_folder_name": "volume_inside_docker"}
        :param image: The image where will execute the command.
        :type: str
        """
        image = image or self.config.get('image_name')
        if not image:
            return {}
        self.pull(image)
        volumes = volumes or self.config.get('volumes')
        env_vars = self.docker_env
        env_vars.update(self.config.get('env_vars'))
        container_env = {key.upper(): value for key, value in env_vars.items()}
        volume_binds = volumes if not working_folder else container.generate_binds(
            volumes, working_folder)
        logger.debug('Executing command: %s', command)
        cont = self.cli2.containers.run(
            image, command, volumes=volume_binds, detach=True, environment=container_env)
        while cont.status != 'exited':
            logger.debug(
                'Waiting for the command to finish, current state: %s', cont.status)
            sleep(5)
            cont.reload()
        cmd_res = cont.logs(stdout=True, stderr=True)
        info = self.inspect(cont.short_id)
        exit_code = info.get('State').get('ExitCode')
        logger.debug("Command result: %s, Exit Code: %s", cmd_res, exit_code)
        cont.remove()
        if exit_code == 0:
            res = {'result': cmd_res.decode()}
        else:
            error = 'Command %(cmd)s failed with exit code: %(code)s %(error)s' % {
                "cmd": command, "code": exit_code, "error": cmd_res.decode()}
            res = {'error': error}
        return res

    def get_env_from_image(self, image=False):
        """Get the environment of a specific image or use the image set in the config,
        then parse the enviroment data in a dictionary.

        :param image: The image's name where get the enviroment.
        :type: str
        :return: return the enviroment parse in a dictionary.
        :rtype: dict
        """
        image = image or self.config.get('image_name')
        info = self.inspect_image(image)
        envs = info.get('Config').get('Env')
        return container.parse_env_vars(envs)



import os
import errno
import logging
import simplejson as json
import deployv_static
from jinja2 import Template
from deployv.helpers import json_helper, utils
from deployv.base.errors import ParameterError, NoItemFound
try:
    from configparser import ConfigParser
except ImportError:
    from ConfigParser import ConfigParser

_logger = logging.getLogger(__name__)


class DeployvConfig:

    def __init__(self, worker_config=None, deploy_config=None, load_defaults=True):
        self._extras = {}
        self.postgres = {}
        self.deployer = {}
        self.http = {}
        self.rmq = {}
        self.instance_config = {}
        self.container_config = {}
        self.maintenance_page = {}
        self.worker_config = self.parse_config(worker_config)
        self._deploy_config = deploy_config or {}
        self.variables = {}
        self.generate_configuration(self._deploy_config, load_defaults=load_defaults)

    def __getattr__(self, name):
        return self._extras.get(name, False)

    def check_config(self, command):
        required_params = {
            'create': ['_deploy_config'],
            'deactivate': ['_deploy_config'],
            'restore': ['_deploy_config'],
            'backup': ['_deploy_config'],
            'deactivate_backup': ['_deploy_config', 'temp_folder', 'prefix'],
            'updatedb': ['_deploy_config'],
            'update_production': ['backup_src', 'compressed_backup_size',
                                  'decompressed_backup_size', 'temp_folder'],
            'build_image': ['_deploy_config'],
        }
        missing_params = []
        if command not in required_params.keys():
            return
        for param in required_params.get(command):
            checker_method = '_check_' + param
            checker_res = True
            value = (getattr(self, param) if hasattr(self, param) else
                     self._extras.get(param, False))
            func = getattr(self, checker_method)
            if func:
                checker_res = func(value)
            if not value or not checker_res:
                missing_params.append(param)
        if missing_params:
            raise ParameterError(
                'Some required parameters are missing, were provided in a wrong'
                ' format or failed a check for the {cmd} command: {params}'
                .format(cmd=command, params=','.join(missing_params)))

    def _check_backup_src(self, backup_src):
        if not os.access(backup_src, os.F_OK | os.R_OK | os.W_OK):
            _logger.error('Unable to write data in the specified backup directory.')
            return False
        return True

    def _check_compressed_backup_size(self, compressed_backup_size):
        parsed_compressed = utils.parse_backup_size(compressed_backup_size)
        if not compressed_backup_size or not parsed_compressed:
            return False
        return True

    def _check_decompressed_backup_size(self, decompressed_backup_size):
        parsed_decompressed = utils.parse_backup_size(decompressed_backup_size)
        if not decompressed_backup_size or not parsed_decompressed:
            return False
        return True

    def _get_backup_src(self, config):
        backup_src = config['group_config'].get('backup_src') or utils.get_backup_src(
            self._deploy_config, self.worker_config)
        return os.path.expanduser(backup_src) if backup_src else False

    def _get_temp_folder(self, config):
        temp_folder = (config['group_config'].get('temp_folder') or
                       self.worker_config.has_option('deployer', 'temp_folder') and
                       self.worker_config.get('deployer', 'temp_folder') or '/tmp')
        res = utils.makedir(temp_folder)
        return res

    def _get_prefix(self, config):
        prefix = config['group_config'].get('prefix') or self._deploy_config.get(
            'instance', {}).get('prefix')
        return prefix

    def _get_compressed_backup_size(self, config):
        compressed_backup_size = config['group_config'].get('compressed_backup_size')
        decompressed_backup_size = config['group_config'].get('decompressed_backup_size')
        parsed_decompressed = utils.parse_backup_size(decompressed_backup_size)
        if not compressed_backup_size and (decompressed_backup_size and parsed_decompressed):
            compressed_value = parsed_decompressed['size_value'] * 0.75
            compressed_unit = parsed_decompressed['size_unit']
            compressed_backup_size = '{size}{unit}'.format(size=compressed_value,
                                                           unit=compressed_unit)
        return compressed_backup_size

    def parse_config(self, additional_config_file=None):
        """Loads and returns the parsed config files in a ConfigParser object.

        The config files are loaded in this order:

        1. /etc/deployv/deployv.conf
        2. /etc/deployv/conf.d/*.conf
        3. ~/.config/deployv/deployv.conf
        4. ~/.config/deployv/conf.d/*.conf
        5. All files received in additional_config_files.

        :param additional_config_file: Additional .conf file to load.
        :type additional_config_files: String

        :returns: The parsed config files.
        :rtype: ConfigParser
        """
        config = self.worker_config or ConfigParser()
        default_config_file = deployv_static.get_template_path('deployv.conf')
        config_files = [default_config_file]
        main_dirs = ['/etc/deployv', os.path.expanduser('~/.config/deployv')]
        for main_dir in main_dirs:
            main_config_file = os.path.join(main_dir, 'deployv.conf')
            if os.access(main_config_file, os.F_OK):
                config_files.append(main_config_file)
            addons_dir = os.path.join(main_dir, 'conf.d')
            if os.path.isdir(addons_dir):
                for filename in os.listdir(addons_dir):
                    filepath = os.path.join(addons_dir, filename)
                    if os.path.isfile(filepath) and filename.endswith('.conf'):
                        config_files.append(filepath)
        if len(config_files) == 1:
            _logger.warning(
                "Warning: no config file detected in the config "
                "paths. You can create a `deployv.conf` file in "
                "`/etc/deployv/` or `~/.config/deployv/` paths. "
                "You can also run `deployvcmd setup_config` to "
                "automatically create the config files for you.")
        if additional_config_file:
            config_files.append(additional_config_file)
        if config_files:
            _logger.info("Loading config files: %s", ", ".join(config_files))
            loaded_files = config.read(config_files)
            if len(config_files) != len(loaded_files):
                _logger.warning("Error loading config files: %s",
                                ", ".join(set(config_files) - set(loaded_files)))
        self.set_worker_config_sections(config)
        return config

    def set_worker_config_sections(self, config):
        """ Sets each section of the worker configuration as an attribute of the config object,
        each attribute will contain a dictionary with the keys of that section. This is so we
        can access to all the configuration in an easier way.

        :param config: Worker configuration
        :type config: ConfigParser object
        """
        booleans = ['use_template', 'use_nginx', 'rmq_ssl', 'pika_logs']
        integers = ['db_port', 'workers', 'max_instances', 'jobs', 'rmq_port', 'rmq_timeout']
        lists = ['extensions']
        if not isinstance(config, ConfigParser):
            raise ParameterError('worker_config: Invalid type {}'.format(type(config)))
        for section in config.sections():
            values = dict(config.items(section))
            try:
                for key in values.keys():
                    if key in booleans:
                        values.update({key: config.getboolean(section, key)})
                    elif key in integers:
                        values.update({key: config.getint(section, key)})
                    elif key in lists:
                        values.update({key: json.loads(config.get(section, key))})
            except ValueError as error:
                error_msg = utils.get_error_message(error)
                _logger.error('Failed to parse the parameter `%s` of the section `%s`, %s',
                              key, section, error_msg)
                raise ParameterError('worker_config > {sect} > {key}: Invalid type {t}'
                                     .format(sect=section, key=key, t=type(values.get(key))))
            setattr(self, section, values)

    def generate_configuration(self, config, load_defaults=True):
        """ Configures the current instance of this class with the provided config dict.
        This method sets the `instance_config` and the  `container_config` attributes
        based on the `instance` and `container_config` keys in the config dict, it
        also adds an attribute for each key in the `group_config` key. Each one of these params
        can have a `_get_{attribute name}` method with special logic to help this method decide
        what value will be stored in that new attribute. Note: These `_get_*` methods **must**
        receive the full config dictionary because some special attributes depend on other values.

        :param config: Configuration that will be used to prepare this object to be used by deployv
        :type config: dict
        :param load_defaults: Wether this method will load the default configuration json or not.
        Default: True
        :type load_defaults: bool
        """
        parsed_config = json_helper.load_json(config)
        if not isinstance(parsed_config, dict):
            raise ParameterError('deploy_config: Invalid type: {tp}'.format(tp=type(config)))
        if load_defaults:
            parsed_config = utils.merge_config(parsed_config)
        res = json.loads(Template(json.dumps(parsed_config)).render(parsed_config))
        # Parameters that require extra logic before setting them
        special_params = ['backup_src', 'temp_folder', 'prefix', 'compressed_backup_size']
        self._deploy_config = res
        self.instance_config = res.get('instance') or {}
        self.container_config = res.get('container_config') or {}
        self.maintenance_page = res.get('maintenance_page') or {}
        # Use a copy instead of the original dictionary so we don't alter the original
        # config, this will allow us to update the properties without the params that
        # were previously set
        extras = res.get('group_config', {}).copy()
        for param in special_params:
            get_method = '_get_' + param
            try:
                func = getattr(self, get_method)
                new_val = func(res)
            except AttributeError:
                raise ParameterError('The attribute {param} was defined as a special'
                                     ' attribute but the expected `_get_{param}()` method was'
                                     ' not found'.format(param=param))
            extras.update({param: new_val})
        self._extras = extras

    def update_configuration(self, worker_config=False, deploy_config=False):
        """ Updates the configuration of the object based on the provided parameters.
        This will help us update the properties of the object in case we modify the
        deploy config or the worker config or we can specify new configurations to
        replace the current ones.

        :param worker_config: Optional worker configuration object, if specified
        it will override the worker_config property and the new properties will
        be generated based on this new config.
        :type worker_config: ConfigParser() object

        :param deploy_config: Optional deploy configuration, if specified
        it will override the _deploy_config property and the new properties will
        be generated based on this new config. This configuration can be the path
        to the json file, the json string or a dictionary.
        :type deploy_config: String
        """
        self.worker_config = worker_config or self.worker_config
        self.set_worker_config_sections(self.worker_config)
        deploy_config = deploy_config or self._deploy_config
        self.generate_configuration(deploy_config)

    def setup_config_dirs(self):
        _logger.info("Writing config files")
        default_config_file = deployv_static.get_template_path('deployv.conf')
        for main_dir in ['/etc/deployv', os.path.expanduser('~/.config/deployv')]:
            config_file_name = os.path.join(main_dir, 'deployv.conf')
            error_msg = "Couldn't write config file: '%s' (skipped)"
            addon_dir = os.path.join(main_dir, 'conf.d')
            is_user_dir = os.path.expanduser('~/') in main_dir
            try:
                os.makedirs(addon_dir)
            except OSError as error:
                if error.errno != errno.EEXIST:
                    _logger.warning(error_msg, config_file_name)
                    continue
            else:
                if is_user_dir:
                    for path in [main_dir.replace('/deployv', ''), main_dir, addon_dir]:
                        self._fix_ownership(path)
            config = ConfigParser()
            config.read([default_config_file, config_file_name])
            try:
                with open(config_file_name, 'w+') as config_file:
                    config.write(config_file)
            except (OSError, IOError):
                _logger.warning(error_msg, config_file_name)
            else:
                _logger.info("Created config file: '%s'", config_file_name)
                if is_user_dir:
                    self._fix_ownership(config_file_name)
        self.worker_config = self.parse_config()

    def _fix_ownership(self, path):
        uid = os.environ.get('SUDO_UID')
        gid = os.environ.get('SUDO_GID')
        if uid is not None:
            os.chown(path, int(uid), int(gid))

    def get_mainrepo_path(self):
        main_repo = utils.get_main_repo(self.instance_config.get('repositories', []))
        if main_repo:
            return main_repo.get('path')
        raise NoItemFound('No main repo found in the config')


def get_config(config_file=None):
    config = DeployvConfig(worker_config=config_file)
    return config.worker_config



# coding: utf-8
""" Base module with all core functionalities needed to help us in the building process of
instances
"""
import locale

try:
    locale.setlocale(locale.LC_ALL, '')
    CHARSET = locale.getlocale(locale.LC_CTYPE)[1]
except locale.Error:
    CHARSET = None
if CHARSET is None:
    CHARSET = 'ascii'



# coding: utf-8

"""This is the core for managing extensions and plugings for the
:class:`~deployv.base.commandv.CommandV` class.

All of the events handlers must inherit from
:class:`~deployv.base.extensions_core.EventListenerBase`.
"""

from functools import wraps
from inspect import signature
import abc
import glob
import imp
import logging
import os
import six
import extend_me


logger = logging.getLogger(__name__)  # pylint: disable=C0103
_eventlistener_meta = extend_me.ExtensibleByHashType._('EventListener', hashattr='name')


def load_extensions():
    """This module just loads all the extensions from the corresponding path.

    :returns: Loaded modules.
    :rtype: dict
    """
    modules = {}
    path = os.path.dirname(os.path.realpath(__file__))
    for path in glob.glob(os.path.join(path, '..', 'extensions/[!_]*.py')):
        name, ext = os.path.splitext(os.path.basename(path))
        modules[name] = imp.load_source(name, path)
    return modules


@six.add_metaclass(_eventlistener_meta)
class EventListenerBase:
    """This is the base class for all the events listener that will be developed in the base
    commands and new ones.

    All inherited classes must have a class Meta with the required information.

    There is a small example in :class:`~deployv.base.extensions_core.TestListener` with the basic
    information configured.
    """

    def __repr__(self):
        try:
            name = self.Meta.name
        except AttributeError:
            name = None

        if name is not None:
            return 'EventListener:%s' % name
        return super(EventListenerBase, self).__repr__()

    @property
    def event(self):

        try:
            event_name = self.Meta.event
        except AttributeError:
            event_name = None
        return event_name

    @property
    def name(self):

        try:
            _name = self.Meta.name
        except AttributeError:
            _name = None
        return _name

    @abc.abstractmethod
    def execute(self, *args, **kwargs):
        """The code that will be executed when the event is *fired* by the
        :class:`~deployv.base.extensions_core.EventManager` according to the event that the class
        registered using the *event* property of the Meta class.

        Check the example :class:`~deployv.base.extensions_core.TestListener` class for more info.

        :returns: The execution result.
        :rtype: dict
        """
        return


class TestListener(EventListenerBase):
    """Just an example plugin to test if plugin logic works.

    Must inherit from :class:`~deployv.base.extensions_core.EventListenerBase`.
    """

    class Meta:
        """This is a test event and the information associated with it.

        The event name and the event listener name that you want to associate with the object (very
        useful for logging and debugging).

        The event name **must** be in the format "before.method.event" or "after.method.event",
        otherwise it won't be executed at all.
        """
        name = 'ListenerName'
        event = 'before.test.event'

    def execute(self, obj):
        logger.debug("Running the event %s", self.Meta.event)


class Singleton(type):

    def __init__(cls, name, bases, p_dict):
        super(Singleton, cls).__init__(name, bases, p_dict)
        cls.instance = None

    def __call__(cls, *args, **kw):
        if cls.instance is None:
            cls.instance = super(Singleton, cls).__call__(*args, **kw)
        return cls.instance


@six.add_metaclass(Singleton)
class EventManager:

    def __init__(self):
        self.__listeners = {}

    def __getitem__(self, name):
        listener = self.__listeners.get(name, False)
        if listener is False:
            try:
                event_cls = type(EventListenerBase).get_class(name)
            except ValueError as error:
                raise KeyError(str(error))

            listener = event_cls()  # pylint: disable=E1102
            self.__listeners[name] = listener
        return listener

    def __getattr__(self, name):
        try:
            return self[name]
        except KeyError:
            raise AttributeError(name)

    def __contains__(self, name):
        return name in self.registered_listeners

    def __dir__(self):
        res = dir(super(EventManager, self))
        res.extend(self.registered_listeners)
        return res

    def fire_event(self, event_name, obj, returned_value=None):
        logger.debug('Firing event %s', event_name)
        res = list()
        for listener in self.registered_listeners:
            if self[listener].event == event_name:
                listener_instance = self[listener]
                logger.debug('Event handler %s', listener_instance.name)
                if event_name.startswith('after.'):
                    lres = listener_instance.execute(obj, returned_value)
                else:
                    lres = listener_instance.execute(obj)
                res.append({listener_instance.name: lres})
        return res

    @property
    def registered_listeners(self):
        return type(EventListenerBase).get_registered_names()

    def refresh(self):
        self.__listeners = {}
        return self


def events(func):
    @wraps(func)
    def func_wrapper(*args, **kwargs):
        obj = signature(func).bind(*args, **kwargs).arguments.get('self')
        pre_event = execute_event('before', func.__name__, obj)
        res = func(*args, **kwargs)
        post_event = execute_event('after', func.__name__, obj, res)
        if pre_event:
            res.update({'before': pre_event})
        if post_event:
            res.update({'after': post_event})
        return res
    return func_wrapper


def execute_event(event, name, obj, args=False):
    event_manager = EventManager()
    logger.debug('Executing %s.events for %s ', event, name)
    return event_manager.fire_event('%s.%s.event' % (event, name), obj, args)



# coding: utf-8
import logging
import pika


_logger = logging.getLogger(__name__)


class ReconnectionException(Exception):
    def __init__(self, message, *args, **kwargs):
        super(ReconnectionException, self).__init__(message, *args, **kwargs)


class RabbitReceiverV:
    """ This is an example consumer that will handle unexpected interactions
    with RabbitMQ such as channel and connection closures.

    If RabbitMQ closes the connection, it will reopen it. You should
    look at the output, as there are limited reasons why the connection may
    be closed, which usually are tied to permission related issues or
    socket timeouts.

    If the channel is closed, it will indicate a problem with one of the
    commands that were issued and that should surface in the output as well.

    """
    EXCHANGE_TYPE = 'topic'

    def __init__(self, listener_configuration, worker_id):
        """ Create a new instance of the consumer class, passing in the AMQP
        URL used to connect to RabbitMQ.

        :param str amqp_url: The AMQP url to connect with

        """
        self.should_reconnect = False
        self._connection = None
        self._channel = None
        self._closing = False
        self._consumer_tag = None
        self._consuming = False

        self._config = listener_configuration
        if self._config.commit_config:
            self._binding_key = '%s.commit' % (self._config.route)
        elif self._config.result_config:
            self._binding_key = self._config.route
        else:
            self._binding_key = '{base}.{wid}'.format(base=self._config.route, wid=worker_id)

    def connect(self):
        """ This method connects to RabbitMQ, returning the connection handle.
        When the connection is established, the on_connection_open method
        will be invoked by pika.

        :rtype: pika.SelectConnection

        """
        _logger.debug('Connecting to %s:%s',
                      self._config.parameters.host,
                      self._config.parameters.port)
        return pika.SelectConnection(
                parameters=self._config.parameters,
                on_open_callback=self.on_connection_open,
                on_open_error_callback=self.on_connection_open_error,
                on_close_callback=self.on_connection_closed)

    def close_connection(self):
        """ This method closes the connection to RabbitMQ."""
        self._consuming = False
        if self._connection.is_closing or self._connection.is_closed:
            _logger.debug('Connection already closed')
            return
        _logger.debug('Closing connection')
        self._connection.close()

    def on_connection_open(self, unused_connection):
        """ This method is called by pika once the connection to RabbitMQ has
        been established. It passes the handle to the connection object in
        case we need it, but in this case, we'll just mark it unused.

        :type unused_connection: pika.SelectConnection

        """
        _logger.debug('Connection opened')
        self.open_channel()

    def on_connection_open_error(self, _unused_connection, err):
        """This method is called by pika if the connection to RabbitMQ
        can't be established.
        :param pika.SelectConnection _unused_connection: The connection
        :param Exception err: The error
        """
        _logger.error('Connection open failed: %s', err)
        self.reconnect()

    def on_connection_closed(self, connection, reason):
        """ This method is invoked by pika when the connection to RabbitMQ is
        closed unexpectedly. Since it is unexpected, we will reconnect to
        RabbitMQ if it disconnects.

        :param pika.connection.Connection connection: The closed connection obj
        :param int reply_code: The server provided reply_code if given
        :param str reply_text: The server provided reply_text if given

        """
        self._channel = None
        if self._closing:
            self._connection.ioloop.stop()
        else:
            _logger.debug('Connection closed: %s, reopening', reason)
            self.reconnect()

    def reconnect(self):
        """ Will be invoked by the IOLoop timer if the connection is
        closed. See the on_connection_closed method.

        """
        # This is the old connection IOLoop instance, stop its ioloop
        self.should_reconnect = True
        self.stop()

    def open_channel(self):
        """ Open a new channel with RabbitMQ by issuing the Channel.Open RPC
        command. When RabbitMQ responds that the channel is open, the
        on_channel_open callback will be invoked by pika.

        """
        _logger.debug('Creating a new channel')
        self._connection.channel(on_open_callback=self.on_channel_open)

    def on_channel_open(self, channel):
        """ This method is invoked by pika when the channel has been opened.
        The channel object is passed in so we can make use of it.

        Since the channel is now open, we'll declare the exchange to use.

        :param pika.channel.Channel channel: The channel object

        """
        _logger.debug('Channel opened')
        self._channel = channel
        self.add_on_channel_close_callback()
        self.setup_exchange(self._config.exchange_name)

    def add_on_channel_close_callback(self):
        """ This method tells pika to call the on_channel_closed method if
        RabbitMQ unexpectedly closes the channel.

        """
        _logger.debug('Adding channel close callback')
        self._channel.add_on_close_callback(self.on_channel_closed)

    def on_channel_closed(self, channel, reason):
        """ Invoked by pika when RabbitMQ unexpectedly closes the channel.
        Channels are usually closed if you attempt to do something that
        violates the protocol, such as re-declare an exchange or queue with
        different parameters. In this case, we'll close the connection
        to shutdown the object.

        :param pika.channel.Channel: The closed channel
        :param int reply_code: The numeric reason the channel was closed
        :param str reply_text: The text reason the channel was closed

        """
        _logger.debug('Channel %i was closed: %s', channel, reason)
        self.close_connection()

    def setup_exchange(self, exchange_name):
        """ Setup the exchange on RabbitMQ by invoking the Exchange.Declare RPC
        command. When it is complete, the on_exchange_declareok method will
        be invoked by pika.

        :param str|unicode exchange_name: The name of the exchange to declare

        """
        _logger.debug('Declaring exchange %s', exchange_name)
        self._channel.exchange_declare(exchange_name,
                                       self.EXCHANGE_TYPE,
                                       callback=self.on_exchange_declareok)

    def on_exchange_declareok(self, unused_frame):
        """ Invoked by pika when RabbitMQ has finished the Exchange.Declare RPC
        command.

        :param pika.Frame.Method unused_frame: Exchange.DeclareOk response frame

        """
        _logger.debug('Exchange declared')
        self.setup_queue(self._binding_key)

    def setup_queue(self, queue_name):
        """ Setup the queue on RabbitMQ by invoking the Queue.Declare RPC
        command. When it is complete, the on_queue_declareok method will
        be invoked by pika.

        :param str|unicode queue_name: The name of the queue to declare.

        """
        _logger.debug('Declaring queue %s', queue_name)
        self._channel.queue_declare(queue_name, callback=self.on_queue_declareok, durable=True)

    def on_queue_declareok(self, method_frame):
        """ Method invoked by pika when the Queue.Declare RPC call made in
        setup_queue has completed. In this method we will bind the queue
        and exchange together with the routing key by issuing the Queue.Bind
        RPC command. When this command is complete, the on_bindok method will
        be invoked by pika.

        :param pika.frame.Method method_frame: The Queue.DeclareOk frame

        """
        _logger.debug('Binding %s to %s with %s',
                      self._config.exchange_name, self._config.queue_name, self._binding_key)
        self._channel.queue_bind(self._binding_key, self._config.exchange_name,
                                 self._binding_key, callback=self.on_bindok)
        _logger.debug('Binding %s to %s with %s',
                      self._config.exchange_name, self._config.queue_name, self._config.route)
        self._channel.queue_bind(self._config.queue_name, self._config.exchange_name,
                                 self._config.route, callback=self.on_bindok)

    def on_bindok(self, unused_frame):
        """ Invoked by pika when the Queue.Bind method has completed. At this
        point we will start consuming messages by calling start_consuming
        which will invoke the needed RPC commands to start the process.

        :param pika.frame.Method unused_frame: The Queue.BindOk response frame

        """
        _logger.debug('Queue bound')
        self.set_qos()

    def set_qos(self):
        """This method sets up the consumer prefetch to only be delivered
        one message at a time. The consumer must acknowledge this message
        before RabbitMQ will deliver another one. You should experiment
        with different prefetch values to achieve desired performance.
        """
        self._channel.basic_qos(prefetch_count=1, callback=self.on_basic_qos_ok)

    def on_basic_qos_ok(self, _unused_frame):
        """Invoked by pika when the Basic.QoS method has completed. At this
        point we will start consuming messages by calling start_consuming
        which will invoke the needed RPC commands to start the process.
        :param pika.frame.Method _unused_frame: The Basic.QosOk response frame
        """
        # TODO: if we need this, remember to set a configurable value
        _logger.debug('QOS set to: 1')
        self.start_consuming()

    def start_consuming(self):
        """ This method sets up the consumer by first calling
        add_on_cancel_callback so that the object is notified if RabbitMQ
        cancels the consumer. It then issues the Basic.Consume RPC command
        which returns the consumer tag that is used to uniquely identify the
        consumer with RabbitMQ. We keep the value to use it when we want to
        cancel consuming.
        """
        _logger.debug('Issuing consumer related RPC commands')
        self.add_on_cancel_callback()
        self._consumer_tag = self._channel.basic_consume(
            self._binding_key, self._callback_funct,
            callback=self.stop_consuming)
        self._consuming = True

    def add_on_cancel_callback(self):
        """ Add a callback that will be invoked if RabbitMQ cancels the consumer
        for some reason. If RabbitMQ does cancel the consumer,
        on_consumer_cancelled will be invoked by pika.

        """
        _logger.debug('Adding consumer cancellation callback')
        self._channel.add_on_cancel_callback(self.on_consumer_cancelled)

    def on_consumer_cancelled(self, method_frame):
        """ Invoked by pika when RabbitMQ sends a Basic.Cancel for a consumer
        receiving messages.

        :param pika.frame.Method method_frame: The Basic.Cancel frame

        """
        _logger.info('Consumer was cancelled remotely, shutting down: %r',
                     method_frame)
        if self._channel:
            self._channel.close()

    def stop_consuming(self, *args, **kwargs):
        """ Tell RabbitMQ that you would like to stop consuming by sending the
        Basic.Cancel RPC command.

        """
        if self._channel:
            _logger.debug('Sending a Basic.Cancel RPC command to RabbitMQ')
            self._channel.basic_cancel(self._consumer_tag, self.on_cancelok)

    def on_cancelok(self, unused_frame):
        """ This method is invoked by pika when RabbitMQ acknowledges the
        cancellation of a consumer. At this point we will close the channel.
        This will invoke the on_channel_closed method once the channel has been
        closed, which will in-turn close the connection.

        :param pika.frame.Method unused_frame: The Basic.CancelOk frame

        """
        self._consuming = False
        _logger.debug('RabbitMQ acknowledged the cancellation of the consumer')
        self.close_channel()

    def close_channel(self):
        """ Call to close the channel with RabbitMQ cleanly by issuing the
        Channel.Close RPC command.

        """
        _logger.debug('Closing the channel')
        self._channel.close()

    def run(self, callback_funct):
        """ Run the example consumer by connecting to RabbitMQ and then
        starting the IOLoop to block and allow the SelectConnection to operate.

        """
        self._callback_funct = callback_funct
        self._connection = self.connect()
        self._connection.ioloop.start()

    def stop(self):
        """ Cleanly shutdown the connection to RabbitMQ by stopping the consumer
        with RabbitMQ. When RabbitMQ confirms the cancellation, on_cancelok
        will be invoked by pika, which will then closing the channel and
        connection. The IOLoop is started again because this method is invoked
        when CTRL-C is pressed raising a KeyboardInterrupt exception. This
        exception stops the IOLoop which needs to be running for pika to
        communicate with RabbitMQ. All of the commands issued prior to starting
        the IOLoop will be buffered but not processed.

        """
        _logger.debug('Stopping')
        if not self._closing:
            self._closing = True
            if self._consuming:
                self.stop_consuming()
                self._connection.ioloop.start()
            else:
                self._connection.ioloop.stop()
        _logger.debug('Stopped')



#!/usr/bin/python
import os
import os.path as osp
import logging
import sys
from deployv.helpers.utils import parse_repo_url
from branchesv.branches import load as load_branches

_logger = logging.getLogger('CloneOCADeps')

TRAVIS_REPO_SLUG = os.environ.get("TRAVIS_REPO_SLUG")
TRAVIS_REPO_OWNER = TRAVIS_REPO_SLUG and TRAVIS_REPO_SLUG.split("/")[0]


def parse_depfile(depfile, version, owner='vauxoo'):
    """ This method, parse plain text file oca_dependencies.txt,
    if you have valid repository name, url and version
    :param depfile: object type open file
    :return: Returns list of tuples [(repo, url, version)]
    """
    deps = []
    for line in depfile:
        line = line.strip()
        if not line or line.startswith('#'):
            continue
        parts = line.split()
        parts_count = len(parts)
        repo = parts[0]
        url = parts_count > 1 and parts[1]
        branch = parts_count > 2 and parts[2]
        commit = parts_count > 3 and parts[3]
        if not branch:
            branch = version
        if not url:
            url = 'https://github.com/%s/%s.git' % (owner, repo)
        if commit == '#':
            # It's not a commit, it's an inline comment
            commit = False
        deps.append((repo, url, branch, commit))
    return deps


def git_checkout(deps_checkout_dir, reponame, url, branch, commit=False,
                 tmp_folder=False, key_path=False):
    """ This method downloads the repositories found in the oca_dependencies
    files and clones them in the specified dir. If he specified directory
    where the dependencies will be cloned already exists
    and is not empty it won't clone the repos
    :param deps_checkout_dir: he directory in which the dependency
    repositories will be cloned
    :param reponame: name for repository
    :param url: url address repository
    :param branch: branch or version repository
    :param commit: Optional commit to clone
    :return: Returns list (full path)
    """
    repo = {
        'branch': branch, 'commit': commit, 'depth': 1, 'name': reponame,
        'path': reponame, 'repo_url': {'origin': url}
    }
    res = load_branches([repo], deps_checkout_dir, tmp_folder, key_path)
    if res:
        return (False, res.get('msg'))
    checkout_dir = osp.join(deps_checkout_dir, reponame)
    return (True, checkout_dir)


def get_dep_filename(deps_checkout_dir, build_dir, file_name,
                     version, tmp_folder, key_path=False):
    """ This method, makes a recursive search directories to get all the
    files listed in the parameters file_name (oca_dependencies.txt)
    :param deps_checkout_dir: he directory in which the dependency
    repositories will be cloned
    :param build_dir: the directory in which the tested
     repositories have been cloned
    :param file_name: filename to be searched
    :return: Returns list with absolute paths
    """
    dependencies = []
    processed = set()
    depfilename = osp.join(build_dir, file_name)
    dependencies.append((depfilename, 'vauxoo'))
    for repo in os.listdir(deps_checkout_dir):
        _logger.debug('examining %s', repo)
        processed.add(repo)
        depfilename = osp.join(deps_checkout_dir, repo, file_name)
        dependencies.append((depfilename, 'vauxoo'))
    for dependency in dependencies:
        try:
            owner = TRAVIS_REPO_OWNER or dependency[1]
            with open(dependency[0]) as depfile:
                deps = parse_depfile(depfile, version, owner=owner)
        except IOError:
            deps = []
        for depname, url, branch, commit in deps:
            _logger.debug('* processing %s', depname)
            if depname in processed:
                continue
            processed.add(depname)
            checkout_dir = git_checkout(
                deps_checkout_dir, depname, url, branch, commit, tmp_folder, key_path)
            if not checkout_dir[0]:
                return checkout_dir
            new_dep = (osp.join(checkout_dir[1], file_name),
                       get_owner_from_url(url))
            if new_dep not in dependencies:
                dependencies.append(new_dep)
    # Return the list of dependencies without the organization so we don't
    # have to modify the methods that receive this result
    res = [dep[0] for dep in dependencies]
    return (True, res)


def get_owner_from_url(url):
    """Parses the specified url in order to get the organization from it.
    This organization will be used to clone all the dependencies of this repo
    that does not have a specific url in the oca_dependencies.txt file. If the
    specified url is not in a correct format, this will return `vauxoo` by
    default.

    :param url: URL of the repo, can be http or ssh
    :type url: str
    :return: The organization of the specified repo (default: vauxoo)
    """
    url_parts = parse_repo_url(url)
    return url_parts.get('namespace') or 'vauxoo'


def run(deps_checkout_dir, build_dir, version, temp_folder, key_path=False):
    """ This method executes above methods
    :param deps_checkout_dir: he directory in which the dependency
    repositories will be cloned
    :param build_dir: the directory in which the tested
    repositories have been cloned
    """
    depens = get_dep_filename(deps_checkout_dir, build_dir, 'oca_dependencies.txt',
                              version, temp_folder, key_path)
    return depens


if __name__ == '__main__':
    run(*sys.argv[1:6])



# coding: utf-8

import logging
import os
from deployv.helpers import utils
try:
    from xmlrpc import client as xmlrpcclient
except ImportError:
    import xmlrpclib as xmlrpcclient


logger = logging.getLogger(__name__)  # pylint: disable=C0103


class NginxV:
    """Wrapper class around supervisor to manage nginx service and site config files.

    :param nginx_working_dir: The path where nginx will store config, logs and pid file.
    :type nginx_working_dir: str
    :param process_name: Process name inside supervisor config file. By default is 'nginx'.
    :type process_name: str
    :param supervisor_url: Url used to stablish communication with supervisor.
        By default is 'unix:///var/run/supervisor.sock'.
    :type supervisor_url: str
    """

    def __init__(self,
                 nginx_working_dir,
                 process_name='nginx',
                 supervisor_url='unix:///var/run/supervisor.sock'):
        self._supervisor_url = supervisor_url
        self._supervisor_server = xmlrpcclient.Server('http://127.0.0.1:9001/RPC2')
        self._process_name = process_name
        self._working_dir = os.path.expanduser(nginx_working_dir)
        logger.debug('Nginx work folder: %s', self._working_dir)
        if not os.path.exists(self._working_dir):
            logger.debug('Creating nginx working path')
            os.mkdir(self._working_dir)

    def restart(self):
        """Restarts the nginx process.

        Generally is only needed when the templates are updated.

        :return: If succeeded or not.
        :rtype: bool
        """
        logger.debug('Sending the HUP signal to nginx')
        res = self._supervisor_server.supervisor.signalProcess(
            self._process_name, 'HUP'
        )
        return res

    def update_sites(self, sites):
        """Updates the sites files with the given options.

        :param sites: Domains and ports in the following format:

            .. code-block:: python

                [
                    {
                        'url': 'sub_domain',
                        'port': instance_port,
                        'lp_port': instance_longpolling_port,
                        'logs': 'path to the instance log'
                    },
                    {
                        'url': 'sub_domain_2',
                        'port': instance_port_2,
                        'lp_port': instalce_longpolling_port_2,
                        'logs': 'path to the instance log'
                    },
                ]

        :type sites: list
        """
        values = {
            'sites': sites,
            'working_dir': self._working_dir
        }
        nginx_config = utils.render_template('nginx_sites.jinja', values)
        logger.debug('Saving generated config to %s', self._working_dir)
        with open(os.path.join(self._working_dir, 'nginx.conf'), 'w') as config:
            config.write(nginx_config)



# coding: utf-8

"""This module substitutes some functions from Gitlab or Github API"""
from os import path
from .utils import parse_repo_url, deploy_key, makedir, get_error_message, clean_files
from branchesv.branches import load as load_branches
from git import Git, exc as exceptions
import logging

_logger = logging.getLogger(__name__)


class GitManager:

    def __init__(self, temp_folder, repos_path=False, key_file=False):
        self.repos_path = repos_path or temp_folder
        self.temp_folder = temp_folder
        self.key_file = key_file

    def get_commit_history(self, repo, ref):
        """ Gets the commit history between the repository in ref repository.
        """
        _logger.debug("Repository: %s", repo)
        _logger.debug("Ref repository: %s", ref)
        config = self._get_config(repo)
        commits = []
        try:
            makedir(config['folder'])
            key_path = deploy_key(self.key_file, self.temp_folder)
            copy_repo = repo.copy()
            copy_repo.update({'path': repo['branch'], 'commit': '', 'depth': False})
            res = load_branches([copy_repo], config['folder'], self.temp_folder, key_path)
            msg_error = res and res.get('msg') or ''
            if not msg_error:
                res, commits = self._get_commits_history(
                    config['checkout_dir'], repo, ref.get('commit', ''))
                msg_error, commits = ('', commits) if res else (commits, [])
        except Exception as error:
            clean_files(config['checkout_dir'])
            msg_error = get_error_message(error)
        history = self._parse_commits_data(commits, repo, ref, msg_error)
        _logger.debug(history)
        return history

    def _get_config(self, repo):
        """Given a working folder returns the .git folder of the repository.

        :param repo: The repository where get the config.
        :type: dict
        :param working_folder: The folder where store the repostiroy.
        :type: str

        :return: return the config needs to get the commit history.
        :rtype: dict
        """
        res = parse_repo_url(repo['repo_url']['origin'])
        folder = path.join(self.repos_path, res['domain'], res['namespace'], res['repo_name'])
        clone_path = path.join(folder, repo['branch'])
        res.update({'folder': folder, 'checkout_dir': clone_path})
        return res

    def _get_commits_history(self, checkout_dir, repo, old_commit):
        """ Gets the commits between old commit and new commit.
        """
        res = []
        new_commit = repo.get('commit')
        gt = Git(checkout_dir)
        try:
            results = gt.log('--pretty=format:"%H|%ad|%an|%ae|%s"', '--date=iso',
                             '%(old)s..%(new)s' % {'old': old_commit, 'new': new_commit})
            lines = results.split('\n') or []
        except exceptions.GitCommandError as error:
            msg_error = get_error_message(error)
            return (False, msg_error)
        for line in lines:
            if not line:
                continue
            fields = line.replace('"', '').split('|')
            res.append({'commit': fields[0], 'date': fields[1], 'author': fields[2],
                        'email': fields[3], 'message': fields[4]})
        return (True, res)

    def _parse_commits_data(self, commits, repo, ref, error):
        repo_url = repo.get('repo_url').get('origin')
        commit_from = ref.get('commit') and ref.get('commit')
        commit_to = repo.get('commit') and repo.get('commit')
        commits = sorted(commits, key=lambda commit: commit['date'])
        return {
            'from': commit_from, 'to': commit_to, 'path': repo.get('path'),
            'branch': repo.get('branch'), 'name': repo.get('name'),
            'repo_url': repo_url, 'commits': commits or [], 'error': error or False,
        }



# coding: utf-8

from __future__ import unicode_literals
import logging
import os
import glob


_logger = logging.getLogger(__name__)


def search_backup(path, customer_id):
    """Searches the best matching backup fot the given client id in te path
    :param path: Path where the backups are stored
    :param customer_id: Unique customer identifier
    :return: Returns the best matching backup (full path)
    """
    _logger.info('Searching backup for %s in "%s"',
                 customer_id, os.path.abspath(path))
    files = sorted(glob.glob(os.path.join(
        path, u'{cid}*'.format(cid=customer_id))), key=lambda s: s.lower())
    _logger.info('Available backups: %s', str(files))
    if files:
        _logger.info('Will use: %s', str(files[-1]))
        return files[-1]
    return None



# coding: utf-8
""" DeployV is a tool ser to help in the process of deploying Odoo instances inside a docker
container, the main idea is to do that remotely and in a unattended way.

For that we created a package with all needed modules (at least those developed by us) to backup,
start, update and build instances.
"""

import logging

__author__ = 'Tulio Ruiz <tulio@vauxoo.com>'
__version__ = '0.9.191'
__all__ = ['base', 'helpers', 'instance', 'messaging']

logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s - %(levelname)-5s - %(name)s.%(funcName)s - %(message)s"
)



# coding: utf-8
import logging
import simplejson as json
import requests
import time
from deployv.messaging.http.httpv import BaseHttpConfiguration
from deployv.messaging.basemsg import BasicMessage

_logger = logging.getLogger('deployv')


class HttpSenderV:
    """ A simple rabbit mq interface just pass the config parameters
    and send the message

    """
    def __init__(self, sender_configuration, sender_id):
        """
        """
        assert isinstance(sender_configuration, BaseHttpConfiguration)
        self._config = sender_configuration.get_result_object()
        self._id = sender_id

    def send_message(self, message):
        retry = 0
        if isinstance(message, BasicMessage):
            data = message.get_message_str()
        else:
            data = json.dumps(message)
        while retry <= 3:
            try:
                res = requests.post(self._config.controller, data=data,
                                    headers=self._config.headers)
                if res.status_code != 200:
                    _logger.error('Error while sending a message: %s', res.status_code)
            except requests.exceptions.ConnectionError:
                retry += 1
                _logger.error('Failed to connect with the instance, retrying %s', retry)
                time.sleep(5)
            break



# coding: utf-8
""" Rabbit and pika wrapper, to have just what we nee and not deal with all
config stuff every time a message is send. The envelope used for this version will be created as
follows specified in the message documentation. A sender is also written with te same purpose

"""

import logging
import pika
import ssl
from deployv.helpers import configuration_helper
try:
    from configparser import ConfigParser
except ImportError:
    from ConfigParser import ConfigParser


_logger = logging.getLogger(__name__)


class BaseRabbitConfiguration:
    """ As the configuration for the sender can be from various sources this will be the base
    class for that, so there is no need to rewrite or change any code from the sender.
    Just need to be sure thar the parameters are fully filled

    :param host: Hostname or ip where the rabbitmq server is hosted
    :param user: Username to login
    :param password: User password
    :param virtual_host: rabbitmq virtual host
    :param queue: The queue prefix to use, the complete name is generated using queue+node_id
    :param exchange: Message exchange to use (mus be the same for all nodes or group)
    :param route: route name where the message will be send ie: test.task.node_id to send
                    a message to all nodes just use test.task
    :param timeout: Connection timeout in seconds
    """
    def __init__(self,
                 host,
                 port,
                 user,
                 password,
                 virtual_host,
                 queue,
                 exchange,
                 route,
                 result=False,
                 timeout=None,
                 use_ssl=True,
                 commit=False):
        self.result_config = result
        self.commit_config = commit
        self.user = user
        self.password = password
        self.credentials = pika.PlainCredentials(self.user, self.password)
        params = {
            'host': host,
            'heartbeat': 0,
            'port': int(port),
            'virtual_host': virtual_host,
            'credentials': self.credentials
        }
        if use_ssl:
            context = ssl.create_default_context()
            context.verify_mode = ssl.CERT_REQUIRED
            context.load_default_certs()
            params.update({'ssl_options': pika.SSLOptions(context)})
        self.parameters = pika.ConnectionParameters(**params)
        self.parameters.socket_timeout = None
        self.queue_name = queue
        self.exchange_name = exchange
        self.route = route
        self.properties = pika.BasicProperties(content_type='application/json',
                                               delivery_mode=2)


class FileRabbitConfiguration(BaseRabbitConfiguration):
    """Reads the configuration from a ini-style file and builds the base
    configuration.

    :param config_object: DeployvConfig instance that contains the configuration read from
        the config files
    :param result: If you need the result configuration or not, this is in the case the workers
        are going to send and ack, result or status message
    """
    def __init__(self, config_object, worker_type=False):
        if isinstance(config_object, ConfigParser):
            self.config = configuration_helper.DeployvConfig(worker_config=config_object)
        elif isinstance(config_object, configuration_helper.DeployvConfig):
            self.config = config_object
        else:
            raise ValueError('No valid config object provided')
        self.config = config_object
        self.wtype = worker_type
        self._config_object = config_object
        user = self.config.rmq.get('rmq_user')
        passwd = self.config.rmq.get('rmq_passwd')
        host = self.config.rmq.get('rmq_server')
        port = self.config.rmq.get('rmq_port')
        vhost = self.config.rmq.get('rmq_vhost')
        timeout = self.config.rmq.get('rmq_timeout')
        exchange = self.config.rmq.get('rmq_exchange')
        use_ssl = self.config.rmq.get('rmq_ssl')
        keys = {'status': ('rmq_stat_queue', 'rmq_status_stat_topic'),
                'result': ('rmq_stat_queue', 'rmq_stat_topic'),
                'commit': ('rmq_task_queue', 'rmq_commit_topic'),
                }
        values = keys.get(worker_type, ('rmq_task_queue', 'rmq_task_topic'))
        queue = self.config.rmq.get(values[0])
        route = self.config.rmq.get(values[1])
        result = worker_type == 'result'
        super(FileRabbitConfiguration, self).__init__(
            host, port, user, passwd, vhost, queue, exchange, route, result, timeout,
            use_ssl, worker_type == 'commit')

    def get_result_object(self):
        return FileRabbitConfiguration(self._config_object, 'result')
